from mwcapture.libmwcapture import mw_notify_status, MW_SUCCEEDED, mw_capture
from pymagewell.exceptions import ProCaptureError

MWCAP_NOTIFY_VIDEO_FRAME_BUFFERING = 0x0100
MWCAP_NOTIFY_VIDEO_FRAME_BUFFERED = 0x0400


class NotificationStatus:
    def __init__(self, mw_notify_status: mw_notify_status):
        self._mw_notify_status = mw_notify_status

    def check_notification_source(self, notification_token: int) -> bool:
        return (self._mw_notify_status.m_ll_notify_status & notification_token) == 0


class Notification:
    def __init__(self, notification_handle: int, channel: int):
        self._handle = notification_handle
        self._channel = channel

    def get_status(self, mw_capture: mw_capture) -> NotificationStatus:
        status: mw_notify_status = mw_notify_status()  # type: ignore
        res: int = mw_capture.mw_get_notify_status(self._channel, self._handle, status)  # type: ignore
        if res != MW_SUCCEEDED:
            raise ProCaptureError("Could not read status of notification")
        return NotificationStatus(status)

    @property
    def handle(self) -> int:
        return self._handle



from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum

from mwcapture.libmwcapture import (
    mw_video_capture_status,
    mw_video_signal_status,
    MWCAP_VIDEO_SIGNAL_NONE,
    MWCAP_VIDEO_SIGNAL_UNSUPPORTED,
    MWCAP_VIDEO_SIGNAL_LOCKING,
    MWCAP_VIDEO_SIGNAL_LOCKED,
    mwcap_video_buffer_info,
    MWCAP_VIDEO_FRAME_STATE_INITIAL,
    MWCAP_VIDEO_FRAME_STATE_BUFFERED,
    MWCAP_VIDEO_FRAME_STATE_F0_BUFFERING,
    MWCAP_VIDEO_FRAME_STATE_F1_BUFFERING,
    mwcap_video_frame_info,
)
from pymagewell.pro_capture_device.device_settings import (
    ImageSizeInPixels,
    AspectRatio,
    ImageCoordinateInPixels,
    DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS,
)


class SignalState(Enum):
    NONE = MWCAP_VIDEO_SIGNAL_NONE
    UNSUPPORTED = MWCAP_VIDEO_SIGNAL_UNSUPPORTED
    LOCKING = MWCAP_VIDEO_SIGNAL_LOCKING
    LOCKED = MWCAP_VIDEO_SIGNAL_LOCKED


@dataclass
class SignalStatus:
    state: SignalState
    start_position: ImageCoordinateInPixels
    image_dimensions: ImageSizeInPixels
    total_dimensions: ImageSizeInPixels
    interlaced: bool
    frame_period_s: float
    aspect_ratio: AspectRatio
    segmented: bool

    @classmethod
    def from_mw_video_signal_status(cls, status: mw_video_signal_status) -> "SignalStatus":
        return SignalStatus(
            state=SignalState(status.state),
            start_position=ImageCoordinateInPixels(col=status.cols, row=status.rows),
            image_dimensions=ImageSizeInPixels(cols=status.cx, rows=status.cy),
            total_dimensions=ImageSizeInPixels(cols=status.cxTotal, rows=status.cyTotal),
            interlaced=bool(status.bInterlaced),
            frame_period_s=float(status.dwFrameDuration * DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS),
            aspect_ratio=AspectRatio(ver=status.nAspectY, hor=status.nAspectX),
            segmented=bool(status.bSegmentedFrame),
        )


@dataclass
class TransferStatus:
    frame_index: int
    whole_frame_transferred: bool
    num_lines_transferred: int
    num_lines_transferred_previously: int

    @classmethod
    def from_mw_video_capture_status(cls, status: mw_video_capture_status) -> "TransferStatus":
        return TransferStatus(
            frame_index=status.iFrame,
            whole_frame_transferred=status.bFrameCompleted,
            num_lines_transferred=status.cyCompleted,
            num_lines_transferred_previously=status.cyCompletedPrev,
        )


@dataclass
class OnDeviceBufferStatus:
    buffer_size_in_frames: int
    """Maximum number of frames in on-board cache"""
    num_chunks_being_buffered: int
    """The number of the slices being bufferred. A frame of video data may contain multiple slices."""
    num_chunks_in_buffer: int
    """The sequence number of slices the latest bufferred piece"""
    buffering_field_index: int
    """The sequence number of fields being bufferred"""
    last_buffered_field_index: int
    """The sequence number of the latest bufferred field"""
    last_buffered_frame_index: int
    """The sequence number of the latest bufferred frame"""
    num_fully_buffered_frames: int
    """Number of fully bufferred full frames"""

    @classmethod
    def from_mwcap_video_buffer_info(cls, info: mwcap_video_buffer_info) -> "OnDeviceBufferStatus":
        return OnDeviceBufferStatus(
            buffer_size_in_frames=info.cMaxFrames,
            num_chunks_being_buffered=info.iNewestBuffering,
            num_chunks_in_buffer=info.iNewestBuffered,
            buffering_field_index=info.iBufferingFieldIndex,
            last_buffered_field_index=info.iBufferedFieldIndex,
            last_buffered_frame_index=info.iNewestBufferedFullFrame,
            num_fully_buffered_frames=info.cBufferedFullFrames,
        )


class FrameState(Enum):
    INTIAL = MWCAP_VIDEO_FRAME_STATE_INITIAL
    BUFFERING_TOP_SUBFRAME = MWCAP_VIDEO_FRAME_STATE_F0_BUFFERING
    BUFFERING_BOTTOM_SUBFRAME = MWCAP_VIDEO_FRAME_STATE_F1_BUFFERING
    BUFFERED = MWCAP_VIDEO_FRAME_STATE_BUFFERED


@dataclass
class DeviceInitTime:
    system_time_as_datetime: datetime
    device_time_in_s: float


def device_time_to_system_time(device_time_in_ticks: int, init_time: DeviceInitTime) -> datetime:
    secs_since_init = device_time_in_ticks * DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS - init_time.device_time_in_s
    return init_time.system_time_as_datetime + timedelta(seconds=secs_since_init)


@dataclass
class FrameInfo:
    state: FrameState
    interlaced: bool
    segmented: bool
    dimensions: ImageSizeInPixels
    aspect_ratio: AspectRatio
    buffering_start_time: datetime
    buffering_complete_time: datetime

    @classmethod
    def from_mwcap_video_frame_info(cls, info: mwcap_video_frame_info, init_time: DeviceInitTime) -> "FrameInfo":

        buffering_start_time_device_ticks = (
            info.allFieldStartTimes[1] if info.bInterlaced else info.allFieldStartTimes[0]
        )
        buffering_complete_time_device_ticks = (
            info.allFieldBufferedTimes[1] if info.bInterlaced else info.allFieldBufferedTimes[0]
        )

        return FrameInfo(
            state=FrameState(info.state),
            interlaced=bool(info.bInterlaced),
            segmented=bool(info.bSegmentedFrame),
            dimensions=ImageSizeInPixels(cols=info.cx, rows=info.cy),
            aspect_ratio=AspectRatio(hor=info.nAspectX, ver=info.nAspectY),
            buffering_start_time=device_time_to_system_time(buffering_start_time_device_ticks, init_time),
            buffering_complete_time=device_time_to_system_time(buffering_complete_time_device_ticks, init_time),
        )



from abc import ABC, abstractmethod
from ctypes import Array, c_char
from dataclasses import dataclass
from datetime import datetime

from pymagewell.events.device_events import (
    TransferCompleteEvent,
    SignalChangeEvent,
    FrameBufferedEvent,
    FrameBufferingEvent,
    TimerEvent,
)
from pymagewell.pro_capture_device.device_settings import (
    ProCaptureSettings,
    TransferMode,
    ImageSizeInPixels,
)
from pymagewell.pro_capture_device.device_status import (
    OnDeviceBufferStatus,
    FrameInfo,
    SignalStatus,
    TransferStatus,
)


@dataclass
class FrameProperties:
    dimensions: ImageSizeInPixels
    size_in_bytes: int


@dataclass
class ProCaptureEvents:
    transfer_complete: TransferCompleteEvent
    """ The event raised by the driver when a transfer (whole frame in normal mode, partial frame in low-latency
    mode) is complete."""
    signal_change: SignalChangeEvent
    """ The event raised by the driver when a source signal change is detected."""
    frame_buffered: FrameBufferedEvent
    """ The event raised by the driver in TransferMode.NORMAL when a frame has been acquired to on-device memory."""
    frame_buffering: FrameBufferingEvent
    """ The event raised by the driver in TransferMode.LOW_LATENCY when a frame has started to be acquired to on-device
    memory."""
    timer_event: TimerEvent
    """ The event raised by the driver in TransferMode.TIMER when a frame has been acquired to on-device memory."""


class ProCaptureDeviceInterface(ABC):
    @abstractmethod
    def __init__(self, settings: ProCaptureSettings):
        raise NotImplementedError()

    @property
    @abstractmethod
    def transfer_mode(self) -> TransferMode:
        raise NotImplementedError()

    @property
    @abstractmethod
    def frame_properties(self) -> FrameProperties:
        raise NotImplementedError()

    @property
    @abstractmethod
    def events(self) -> ProCaptureEvents:
        raise NotImplementedError()

    @abstractmethod
    def schedule_timer_event(self) -> None:
        raise NotImplementedError()

    @property
    @abstractmethod
    def buffer_status(self) -> OnDeviceBufferStatus:
        raise NotImplementedError()

    @property
    @abstractmethod
    def frame_info(self) -> FrameInfo:
        raise NotImplementedError()

    @property
    @abstractmethod
    def signal_status(self) -> SignalStatus:
        raise NotImplementedError()

    @property
    @abstractmethod
    def transfer_status(self) -> TransferStatus:
        raise NotImplementedError()

    @abstractmethod
    def start_grabbing(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def stop_grabbing(self) -> None:
        raise NotImplementedError()

    @property
    @abstractmethod
    def fps(self) -> float:
        raise NotImplementedError()

    @abstractmethod
    def start_a_frame_transfer(self, frame_buffer: Array[c_char]) -> datetime:
        raise NotImplementedError()

    @abstractmethod
    def shutdown(self) -> None:
        raise NotImplementedError()



import logging
from ctypes import create_unicode_buffer, Array, c_char, addressof
from datetime import datetime, timedelta
from typing import cast, Optional

from mwcapture.libmwcapture import (
    mw_capture,
    mwcap_video_buffer_info,
    mwcap_video_frame_info,
    mw_video_signal_status,
    MW_SUCCEEDED,
    mw_video_capture_status,
    mw_device_time,
    MWCAP_VIDEO_DEINTERLACE_BLEND,
    MWCAP_VIDEO_ASPECT_RATIO_CROPPING,
    MWCAP_VIDEO_COLOR_FORMAT_UNKNOWN,
    MWCAP_VIDEO_QUANTIZATION_UNKNOWN,
    MWCAP_VIDEO_SATURATION_UNKNOWN,
)
from pymagewell.pro_capture_device.device_status import (
    TransferStatus,
    SignalStatus,
    SignalState,
    OnDeviceBufferStatus,
    FrameInfo,
    DeviceInitTime,
)
from pymagewell.events.device_events import (
    RegisterableEvent,
    SignalChangeEvent,
    FrameBufferingEvent,
    FrameBufferedEvent,
    FrameTransferCompleteEvent,
    TransferCompleteEvent,
    PartialFrameTransferCompleteEvent,
    TimerEvent,
)
from pymagewell.events.notification import Notification
from pymagewell.pro_capture_device.device_settings import (
    TransferMode,
    ProCaptureSettings,
    DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS,
)
from pymagewell.pro_capture_device.pro_capture_device_impl import ProCaptureDeviceImpl
from pymagewell.pro_capture_device.device_interface import ProCaptureEvents
from pymagewell.exceptions import ProCaptureError


logger = logging.getLogger(__name__)


class ProCaptureDevice(ProCaptureDeviceImpl, mw_capture):
    """`ProCaptureDevice` represents a physical Magewell ProCapture device. It is intended to be controlled by a
    `ProCaptureDeviceController`.

    ProCaptureDevice configures and initiates frame acquisition and transfer. It is responsible for constructing and
    registering events with the Magewell driver. It also provides methods for accessing information about the video
    source connected to the device. ProCaptureDevice inherits from the `mw_capture` class provided by Magwell's
    Windows SDK.
    """

    def __init__(self, settings: ProCaptureSettings):
        """
        Args:
            settings: A `ProCaptureSettings` object containing the settings for the device.
        """
        ProCaptureDeviceImpl.__init__(self, settings)
        mw_capture.__init__(self)  # type: ignore

        self.mw_capture_init_instance()  # type: ignore
        self.mw_refresh_device()  # type: ignore
        self._channel = _create_channel(self)

        self._device_init_time = DeviceInitTime(
            device_time_in_s=self._get_device_time_in_s(), system_time_as_datetime=datetime.now()
        )

        self._timer = _FrameTimer(self, self._channel, self._register_timer_event(TimerEvent()))

        self._signal_change_event = cast(SignalChangeEvent, self._register_event(SignalChangeEvent()))

        self._frame_buffered_event = FrameBufferedEvent()
        self._frame_buffering_event = FrameBufferingEvent()

        if self._settings.transfer_mode == TransferMode.NORMAL:
            self._frame_buffered_event = cast(FrameBufferedEvent, self._register_event(self._frame_buffered_event))
            self._transfer_complete_event: TransferCompleteEvent = FrameTransferCompleteEvent()

        elif self._settings.transfer_mode == TransferMode.LOW_LATENCY:
            self._frame_buffering_event = cast(FrameBufferingEvent, self._register_event(self._frame_buffering_event))
            self._transfer_complete_event = PartialFrameTransferCompleteEvent()

        elif self._settings.transfer_mode.TIMER:
            self._transfer_complete_event = FrameTransferCompleteEvent()

    @property
    def events(self) -> ProCaptureEvents:
        """events property
        Returns:
            A `ProCaptureEvents` object containing handles to the events generated by the device during frame grabbing.
        """
        return ProCaptureEvents(
            transfer_complete=self._transfer_complete_event,
            signal_change=self._signal_change_event,
            frame_buffered=self._frame_buffered_event,
            frame_buffering=self._frame_buffering_event,
            timer_event=self._timer.event,
        )

    def _register_event(self, event: RegisterableEvent) -> RegisterableEvent:
        notification_handle = self.mw_register_notify(self._channel, event.win32_event, event.registration_token)  # type: ignore
        event.register(Notification(notification_handle, self._channel))
        return event

    def _register_timer_event(self, event: TimerEvent) -> TimerEvent:
        """The _FrameTimer class handles constructing TimerEvents and registering them here."""
        notification_handle = self.mw_register_timer(self._channel, event.win32_event)  # type: ignore
        event.register(Notification(notification_handle, self._channel))
        return event

    def schedule_timer_event(self) -> None:
        """For use if the TransferMode is TransferMode.TIMER, this method schedules a timer event to trigger the transfer of the
        next frame."""
        self._timer.schedule_timer_event(self._get_device_time_in_ticks())

    @property
    def buffer_status(self) -> OnDeviceBufferStatus:
        """buffer_status property
        Returns:
            A `OnDeviceBufferStatus` object containing information about the device's buffer.
        """
        buffer_info = mwcap_video_buffer_info()
        self.mw_get_video_buffer_info(self._channel, buffer_info)  # type: ignore
        return OnDeviceBufferStatus.from_mwcap_video_buffer_info(buffer_info)

    @property
    def frame_info(self) -> FrameInfo:
        """frame_info property
        Returns:
            A `FrameInfo` object containing information about the most recent frame.
        """
        frame_info = mwcap_video_frame_info()
        self.mw_get_video_frame_info(self._channel, self.buffer_status.last_buffered_frame_index, frame_info)  # type: ignore
        return FrameInfo.from_mwcap_video_frame_info(frame_info, self._device_init_time)

    @property
    def signal_status(self) -> SignalStatus:
        """signal_status property
        Returns:
            A `SignalStatus` object containing information about the signal connected to the device.
        """
        mw_signal_status = mw_video_signal_status()
        self.mw_get_video_signal_status(self._channel, mw_signal_status)  # type: ignore
        return SignalStatus.from_mw_video_signal_status(mw_signal_status)

    def start_grabbing(self) -> None:
        """Starts the hardware acquiring frames."""
        start_capture_result = self.mw_start_video_capture(self._channel, self.events.transfer_complete.win32_event)  # type: ignore
        if start_capture_result != MW_SUCCEEDED:
            raise ProCaptureError(f"Start capture failed (error code {start_capture_result}).")
        if self.signal_status.state == SignalState.NONE:
            logger.warning("Input signal status: None")
        elif self.signal_status.state == SignalState.UNSUPPORTED:
            logger.warning("Input signal status: Unsupported")
        elif self.signal_status.state == SignalState.LOCKING:
            logger.info("Input signal status: Locking")
        elif self.signal_status.state == SignalState.LOCKED:
            logger.info("Input signal status: Locked")

        # Exit if signal not locked
        if self.signal_status.state != SignalState.LOCKED:
            self.mw_stop_video_capture(self._channel)  # type: ignore
            self.shutdown()
            raise ProCaptureError("Signal not locked. It is likely that no video signal is present.")

    def stop_grabbing(self) -> None:
        """Stops the hardware acquiring frames."""
        self.mw_stop_video_capture(self._channel)  # type: ignore

    @property
    def transfer_status(self) -> TransferStatus:
        """transfer_status property
        Returns:
            A `TransferStatus` object containing information about the transfer of frames from the device to the PC,
            including the number of frames transferred, and the number of lines of the current frame that have been
            transferred.
        """
        mw_capture_status = mw_video_capture_status()
        self.mw_get_video_capture_status(self._channel, mw_capture_status)  # type: ignore
        return TransferStatus.from_mw_video_capture_status(mw_capture_status)

    def _get_device_time_in_ticks(self) -> mw_device_time:
        """Read a timestamp from the device."""
        time = mw_device_time()  # type: ignore
        result = self.mw_get_device_time(self._channel, time)  # type: ignore
        if result != MW_SUCCEEDED:
            raise ProCaptureError("Failed to read time from device")
        else:
            return time

    def _get_device_time_in_s(self) -> float:
        return int(self._get_device_time_in_ticks().m_ll_device_time.value) * DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS

    def start_a_frame_transfer(self, frame_buffer: Array[c_char]) -> datetime:
        """start_a_frame_transfer starts the transfer of lines from the device to a buffer in PC memory.
        Args:
            frame_buffer (Array[c_char]): A buffer in PC memory to which the lines of the current frame will be
            transferred.
        Returns:
            The time (datetime.datetime) at which the transfer of the first line of the current frame started.
        """
        in_low_latency_mode = self.transfer_mode == TransferMode.LOW_LATENCY
        notify_size = self._settings.num_lines_per_chunk if in_low_latency_mode else 0

        seconds_since_init = self._get_device_time_in_s() - self._device_init_time.device_time_in_s
        frame_timestamp = self._device_init_time.system_time_as_datetime + timedelta(seconds=seconds_since_init)
        result = self.mw_capture_video_frame_to_virtual_address_ex(  # type: ignore
            hchannel=self._channel,
            iframe=self.buffer_status.last_buffered_frame_index,
            pbframe=addressof(frame_buffer),
            cbframe=self._settings.image_size_in_bytes,
            cbstride=self._settings.min_stride,
            bbottomup=False,  # this is True in the C++ example, but false in python example,
            pvcontext=0,
            dwfourcc=self._settings.color_format.value,  # color format of captured frames
            cx=self._settings.dimensions.cols,
            cy=self._settings.dimensions.rows,
            dwprocessswitchs=0,
            cypartialnotify=notify_size,
            hosdimage=0,
            posdrects=0,
            cosdrects=0,
            scontrast=100,
            sbrightness=0,
            ssaturation=100,
            shue=0,
            deinterlacemode=MWCAP_VIDEO_DEINTERLACE_BLEND,
            aspectratioconvertmode=MWCAP_VIDEO_ASPECT_RATIO_CROPPING,
            prectsrc=0,  # 0 in C++ example, but configured using CLIP settings in python example,
            prectdest=0,
            naspectx=0,
            naspecty=0,
            colorformat=MWCAP_VIDEO_COLOR_FORMAT_UNKNOWN,
            quantrange=MWCAP_VIDEO_QUANTIZATION_UNKNOWN,
            satrange=MWCAP_VIDEO_SATURATION_UNKNOWN,
        )
        if result != MW_SUCCEEDED:
            raise ProCaptureError(f"Frame grab failed with error code {result}")
        else:
            return frame_timestamp

    def shutdown(self) -> None:
        """shutdown releases the hardware resources used by the device."""
        self._timer.shutdown()
        self._signal_change_event.destroy()
        self._transfer_complete_event.destroy()
        self._frame_buffered_event.destroy()
        self._frame_buffering_event.destroy()

        self.mw_close_channel(self._channel)  # type: ignore
        self.mw_capture_exit_instance()  # type: ignore


def _create_channel(capturer: mw_capture) -> int:
    t_path = create_unicode_buffer(128)
    capturer.mw_get_device_path(0, t_path)  # type: ignore
    return capturer.mw_open_channel_by_path(t_path)  # type: ignore


class _FrameTimer:
    """If the devices transfer mode is "Timer", this class is used to generate events triggering the transfer of frames
    from the device."""

    def __init__(self, device: ProCaptureDevice, channel: int, event: TimerEvent):
        self._device = device
        self._channel = channel
        self._frame_expire_time: Optional[mw_device_time] = None
        self._timer_event = event

    @property
    def event(self) -> TimerEvent:
        return self._timer_event

    def schedule_timer_event(self, device_time_now: mw_device_time) -> None:

        if self._frame_expire_time is None:
            self._frame_expire_time = device_time_now
        self._frame_expire_time.m_ll_device_time.value += int(
            self._device.signal_status.frame_period_s / DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS
        )

        if self._timer_event.is_registered:
            result = self._device.mw_schedule_timer(
                self._channel,
                self._timer_event.notification.handle,  # type: ignore
                self._frame_expire_time.m_ll_device_time,
            )
        else:
            raise ProCaptureError("Timer event not registered with device.")
        if result != MW_SUCCEEDED:
            raise ProCaptureError("Failed to schedule frame timer")

    def shutdown(self) -> None:
        self._timer_event.destroy()



import logging
from typing import Any

logger = logging.getLogger(__name__)

logger.warning("win32api and win32 event packages not available. pymagewell will only work with MockProCaptureDevice.")

WAIT_OBJECT_0 = 0


def CreateEvent(a: Any, b: Any, c: Any, d: Any) -> Any:
    pass


def SetEvent(a: Any) -> None:
    pass


def WaitForMultipleObjects(a: Any, b: Any, c: Any) -> Any:
    return WAIT_OBJECT_0 + 0


def WaitForSingleObject(a: Any, b: Any) -> Any:
    pass



from mwcapture.libmwcapture import MWCAP_NOTIFY_VIDEO_SIGNAL_CHANGE
from pymagewell.events.event_base import Event, RegisterableEvent
from pymagewell.events.notification import (
    MWCAP_NOTIFY_VIDEO_FRAME_BUFFERING,
    MWCAP_NOTIFY_VIDEO_FRAME_BUFFERED,
)


class SignalChangeEvent(RegisterableEvent):
    """This event is raised by the driver if the source video signal changes"""

    @property
    def registration_token(self) -> int:
        return MWCAP_NOTIFY_VIDEO_SIGNAL_CHANGE


class FrameBufferingEvent(RegisterableEvent):
    """Raised when frame starts to be acquired by the hardware."""

    @property
    def registration_token(self) -> int:
        return MWCAP_NOTIFY_VIDEO_FRAME_BUFFERING


class FrameBufferedEvent(RegisterableEvent):
    """Raised when a frame has been completely acquired by the hardware."""

    @property
    def registration_token(self) -> int:
        return MWCAP_NOTIFY_VIDEO_FRAME_BUFFERED


class TimerEvent(RegisterableEvent):
    """The driver provides a timer function to acquire frames at a custom rate. The _FrameTimer configures the driver
    to raise this event when the timer ticks/"""

    @property
    def registration_token(self) -> None:
        return None


class TransferCompleteEvent(Event):
    pass


class FrameTransferCompleteEvent(TransferCompleteEvent):
    """This event intended to be provided to Device.mw_start_video_capture(). The driver will then raise this event
    once a capture is complete. It is only used in normal mode, to signify that a whole frame has been transferred
    to PC memory."""

    pass


class PartialFrameTransferCompleteEvent(TransferCompleteEvent):
    """This event is intended to be provided to Device.mw_start_video_capture(). The driver will then raise this event
    once a capture is complete. It is only used in low-latency mode, and means that `cypartialnotify` lines have been
    transferred to PC memory."""

    pass



# Christian Baker, King's College London
# Copyright (c) 2022 School of Biomedical Engineering & Imaging Sciences, King's College London
# Licensed under the MIT. You may obtain a copy at https://opensource.org/licenses/MIT.
"""
`pymagewell` is a Python library for interfacing with Magewell ProCapture frame grabbers.

It is based on (and includes) Magewell's Windows SDK and is therefore Windows only. However, it provides a mock class
that for testing and development that does not depend on the SDK's Windows .dll files, so `pymagwell` can also be
installed on macOS and Linux.

* [GitHub page](https://github.com/KCL-BMEIS/pymagewell)
* [API reference documentation](https://kcl-bmeis.github.io/pymagewell/)
* [PyPI page](https://pypi.org/project/pymagewell/)

### Installation

```bash
pip install pymagewell
```

### Example of use

A full working example is provided in
[`example_script.py`](https://github.com/KCL-BMEIS/pymagewell/blob/main/example_script.py).

"""
from . import _version
from .pro_capture_controller import ProCaptureController

from .pro_capture_device import ProCaptureDevice
from .pro_capture_device.device_settings import ProCaptureSettings, TransferMode, ImageSizeInPixels, ColourFormat
from .pro_capture_device.mock_pro_capture_device import MockProCaptureDevice

__version__ = _version.get_versions()["version"]  # type: ignore

__all__ = [
    "ProCaptureDevice",
    "MockProCaptureDevice",
    "ProCaptureController",
    "ProCaptureSettings",
    "TransferMode",
    "ImageSizeInPixels",
    "ColourFormat",
]



import math
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import cast

from mwcapture.libmwcapture import (
    fourcc_calc_min_stride,
    MWFOURCC_NV12,
    fourcc_calc_image_size,
    MWFOURCC_BGR24,
    mwcap_smpte_timecode,
    MWFOURCC_UNK,
    MWFOURCC_GREY,
    MWFOURCC_Y800,
    MWFOURCC_BGRA,
    MWFOURCC_Y8,
    MWFOURCC_Y16,
    MWFOURCC_RGB15,
    MWFOURCC_RGB16,
    MWFOURCC_RGB24,
    MWFOURCC_RGBA,
    MWFOURCC_ARGB,
    MWFOURCC_BGR15,
    MWFOURCC_BGR16,
    MWFOURCC_ABGR,
    MWFOURCC_NV16,
    MWFOURCC_NV61,
    MWFOURCC_I422,
    MWFOURCC_YV16,
    MWFOURCC_YUY2,
    MWFOURCC_YUYV,
    MWFOURCC_UYVY,
    MWFOURCC_BGR10,
    MWFOURCC_RGB10,
    MWFOURCC_V410,
    MWFOURCC_Y410,
    MWFOURCC_V210,
    MWFOURCC_V408,
    MWFOURCC_UYVA,
    MWFOURCC_AYUV,
    MWFOURCC_V308,
    MWFOURCC_IYU2,
    MWFOURCC_P210,
    MWFOURCC_P010,
    MWFOURCC_NV21,
    MWFOURCC_YV12,
    MWFOURCC_IYUV,
    MWFOURCC_I420,
    MWFOURCC_VYUY,
    MWFOURCC_YVYU,
    MWFOURCC_VYUA,
)

DEVICE_CLOCK_TICK_PERIOD_IN_SECONDS = 1e-7


class ColourFormat(Enum):
    """ Enumeration of the supported colour formats. """
    UNK = MWFOURCC_UNK
    GREY = MWFOURCC_GREY
    Y800 = MWFOURCC_Y800
    Y8 = MWFOURCC_Y8
    Y16 = MWFOURCC_Y16
    RGB15 = MWFOURCC_RGB15
    RGB16 = MWFOURCC_RGB16
    RGB24 = MWFOURCC_RGB24
    RGBA = MWFOURCC_RGBA
    ARGB = MWFOURCC_ARGB
    BGR15 = MWFOURCC_BGR15
    BGR16 = MWFOURCC_BGR16
    BGR24 = MWFOURCC_BGR24
    BGRA = MWFOURCC_BGRA
    ABGR = MWFOURCC_ABGR
    MNV16 = MWFOURCC_NV16
    NV61 = MWFOURCC_NV61
    I422 = MWFOURCC_I422
    YV16 = MWFOURCC_YV16
    YUY2 = MWFOURCC_YUY2
    YUYV = MWFOURCC_YUYV
    UYVY = MWFOURCC_UYVY
    YVYU = MWFOURCC_YVYU
    VYUY = MWFOURCC_VYUY
    I420 = MWFOURCC_I420
    IYUV = MWFOURCC_IYUV
    NV12 = MWFOURCC_NV12
    YV12 = MWFOURCC_YV12
    NV21 = MWFOURCC_NV21
    P010 = MWFOURCC_P010
    P210 = MWFOURCC_P210
    IYU2 = MWFOURCC_IYU2
    V308 = MWFOURCC_V308
    AYUV = MWFOURCC_AYUV
    UYVA = MWFOURCC_UYVA
    V408 = MWFOURCC_V408
    VYUA = MWFOURCC_VYUA
    V210 = MWFOURCC_V210
    Y410 = MWFOURCC_Y410
    V410 = MWFOURCC_V410
    RGB10 = MWFOURCC_RGB10
    BGR10 = MWFOURCC_BGR10


@dataclass
class ImageCoordinateInPixels:
    row: int
    col: int


@dataclass
class ImageSizeInPixels:
    cols: int
    rows: int


@dataclass
class AspectRatio:
    hor: int
    ver: int


class TransferMode(Enum):
    """ Enumeration of the supported methods for triggering the transfer of frames to the PC. """
    TIMER = 0
    """ Transferred are triggered by a software timer event, allowing arbitrary frame rates. This is the only mode
        supported by MockProCaptureDevice. """
    NORMAL = 1
    """ Transfers are triggered by a notification received from the device when a whole frame has been received, and
    therefore grabbing happens at the source frame rate."""
    LOW_LATENCY = 2
    """ Transfers are triggered by a notification received from the device when the first chunk of a frame has been
    received. Grabbing happens at the source frame rate, but with a lower latency."""


@dataclass
class FrameTimeCode:
    hour: int
    minute: int
    second: int
    frame: int

    @classmethod
    def from_mwcap_smpte_timecode(cls, tc: mwcap_smpte_timecode) -> "FrameTimeCode":
        return FrameTimeCode(
            hour=int(tc[0]),  # type: ignore
            minute=int(tc[1]),  # type: ignore
            second=int(tc[2]),  # type: ignore
            frame=int(tc[3]),  # type: ignore
        )

    @classmethod
    def now(cls, frame_period_s: float) -> "FrameTimeCode":
        microseconds_per_frame = frame_period_s * 1e6
        now = datetime.now()
        return FrameTimeCode(
            hour=now.hour,
            minute=now.minute,
            second=now.second,
            frame=int(round(now.microsecond / microseconds_per_frame)),
        )

    def as_datetime(self, frame_period_s: float) -> datetime:
        microseconds_per_frame = frame_period_s * 1e6
        now = datetime.now()
        return datetime(
            year=now.year,
            month=now.month,
            day=now.day,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=int(self.frame * microseconds_per_frame),
        )


@dataclass
class ProCaptureSettings:
    """ Settings for the ProCapture device. """
    dimensions: ImageSizeInPixels = ImageSizeInPixels(1920, 1080)
    """The dimensions of the frames to be acquired in pixels."""
    color_format: ColourFormat = ColourFormat.BGR24
    """The colour format of the frames to be acquired."""
    transfer_mode: TransferMode = TransferMode.NORMAL
    """The method to use for transferring frames from the device to the PC. See `TransferMode` for details."""
    num_lines_per_chunk: int = 64
    """The number of lines of a frame to transfer at a time (for `TransferMode.LOW_LATENCY` transfers)."""

    def __post_init__(self) -> None:
        _check_valid_chunk_size(self.num_lines_per_chunk)

    @property
    def min_stride(self) -> int:
        return cast(int, fourcc_calc_min_stride(self.color_format.value, self.dimensions.cols, 2))  # type: ignore

    @property
    def image_size_in_bytes(self) -> int:
        if self.color_format == ColourFormat.NV12:
            return self.dimensions.cols * self.dimensions.rows * 2  # copied from line 223 of capture.py
        else:
            return cast(
                int,
                fourcc_calc_image_size(  # type: ignore
                    self.color_format,
                    self.dimensions.cols,
                    self.dimensions.rows,
                    self.min_stride,
                ),
            )


def _check_valid_chunk_size(n_lines_per_chunk: int) -> None:
    if n_lines_per_chunk < 64:
        raise ValueError("Minimum number of lines per chunk is 64.")
    elif round(math.log2(n_lines_per_chunk)) != math.log2(n_lines_per_chunk):
        raise ValueError("Number of lines per chunk must be a power of 2.")



from threading import Thread
from time import sleep, perf_counter
from unittest import TestCase

from pymagewell.events.event_base import Event, wait_for_event, wait_for_events
from pymagewell.events.device_events import TransferCompleteEvent, TimerEvent, FrameBufferedEvent
from pymagewell.events.notification import Notification


class TestEvents(TestCase):
    def setUp(self) -> None:
        pass

    def tearDown(self) -> None:
        pass

    def test_wait_for_event(self) -> None:
        event = Event()
        event_thread = Thread(target=_set_event_after_100ms, args=[event])
        start_time = perf_counter()
        event_thread.start()
        wait_for_event(event, timeout_ms=1000)
        time_waited = perf_counter() - start_time
        self.assertTrue((time_waited > 0.1) and (time_waited < 0.15))

    def test_wait_for_events(self) -> None:
        events = [TransferCompleteEvent(), TimerEvent()]
        event_thread = Thread(target=_set_event_after_100ms, args=[events[0]])
        start_time = perf_counter()
        event_thread.start()
        event_that_occurred = wait_for_events(events, timeout_ms=1000)
        time_waited = perf_counter() - start_time
        self.assertTrue((time_waited > 0.1) and (time_waited < 0.15))
        self.assertTrue(isinstance(event_that_occurred, TransferCompleteEvent))

    def test_register_event(self) -> None:
        event = FrameBufferedEvent()
        self.assertFalse(event.is_registered)
        event.register(Notification(0, 0))
        self.assertTrue(event.is_registered)


def _set_event_after_100ms(event: Event) -> None:
    sleep(0.1)
    event.set()



from datetime import datetime
from time import perf_counter
from unittest import TestCase

import pytest
from numpy import diff, array

from pymagewell.pro_capture_controller import ProCaptureController
from pymagewell.pro_capture_device import ProCaptureDevice
from pymagewell.pro_capture_device.device_interface import ProCaptureDeviceInterface
from pymagewell.pro_capture_device.device_settings import ProCaptureSettings, TransferMode
from pymagewell.pro_capture_device.mock_pro_capture_device import MockProCaptureDevice


@pytest.mark.usefixtures("hardware_mode")
class TestCaptureController(TestCase):
    def setUp(self) -> None:
        device_settings = ProCaptureSettings()
        device_settings.transfer_mode = TransferMode.TIMER
        assert hasattr(self, "hardware_mode_is_set")
        if self.hardware_mode_is_set:  # type: ignore
            self._device: ProCaptureDeviceInterface = ProCaptureDevice(device_settings)
        else:
            self._device = MockProCaptureDevice(device_settings)

        self._controller = ProCaptureController(device=self._device)

    def tearDown(self) -> None:
        self._controller.shutdown()

    def test_buffering_timestamps(self) -> None:
        frame = self._controller.transfer_when_ready(timeout_ms=1000)

        buffering_duration = (frame.timestamps.buffering_complete - frame.timestamps.buffering_started).total_seconds()
        self.assertTrue(0 <= buffering_duration <= 0.1)

        seconds_since_buffering_started = (datetime.now() - frame.timestamps.buffering_started).total_seconds()
        self.assertTrue(0 <= seconds_since_buffering_started < 0.25)

        seconds_since_buffering_complete = (datetime.now() - frame.timestamps.buffering_complete).total_seconds()
        self.assertTrue(0 <= seconds_since_buffering_complete < 0.25)

    def test_transfer_timestamps(self) -> None:
        frame = self._controller.transfer_when_ready(timeout_ms=1000)

        transfer_duration = (frame.timestamps.transfer_complete - frame.timestamps.transfer_started).total_seconds()
        self.assertTrue(0 <= transfer_duration <= 0.1)

        seconds_since_transfer_started = (datetime.now() - frame.timestamps.transfer_started).total_seconds()
        self.assertTrue(0 <= seconds_since_transfer_started < 0.25)

        seconds_since_transfer_complete = (datetime.now() - frame.timestamps.transfer_complete).total_seconds()
        self.assertTrue(0 <= seconds_since_transfer_complete < 0.25)

    def test_frame_size(self) -> None:
        frame = self._controller.transfer_when_ready(timeout_ms=1000)
        self.assertEqual(frame.dimensions.rows, self._device.frame_properties.dimensions.rows)
        self.assertEqual(frame.dimensions.cols, self._device.frame_properties.dimensions.cols)

    def test_frame_period(self) -> None:
        times_frames_received = []
        for _ in range(10):
            _ = self._controller.transfer_when_ready(timeout_ms=1000)
            times_frames_received.append(perf_counter())
        mean_frame_period = diff(array(times_frames_received)).mean()
        self.assertAlmostEqual(self._device.signal_status.frame_period_s, mean_frame_period, 1)



# Christian Baker, King's College London
# Copyright (c) 2022 School of Biomedical Engineering & Imaging Sciences, King's College London
# Licensed under the MIT. You may obtain a copy at https://opensource.org/licenses/MIT.

""" pymagewell provides a high-level python interface to the Magewell ProCapture SDK """



from abc import ABC

from pymagewell.pro_capture_device.device_interface import (
    ProCaptureDeviceInterface,
    FrameProperties,
)
from pymagewell.pro_capture_device.device_settings import (
    TransferMode,
    ProCaptureSettings,
)


class ProCaptureDeviceImpl(ProCaptureDeviceInterface, ABC):
    """Implements methods of the interface that do not require communication with the driver or hardware, or require use
    of the Events system, and therefore are common to Mock and Real implementations of the interface."""

    def __init__(self, settings: ProCaptureSettings):
        self._settings = settings

    @property
    def transfer_mode(self) -> TransferMode:
        return self._settings.transfer_mode

    @property
    def frame_properties(self) -> FrameProperties:
        return FrameProperties(dimensions=self._settings.dimensions, size_in_bytes=self._settings.image_size_in_bytes)

    @property
    def fps(self) -> float:
        if self.signal_status.interlaced:
            fps = 2e7 / self.signal_status.frame_period_s
        else:
            fps = 1e7 / self.signal_status.frame_period_s
        return fps



import logging
from ctypes import Array, c_char
from datetime import datetime
from operator import mod
from threading import Thread
from time import monotonic, sleep
from typing import Optional

from cv2 import circle, FONT_HERSHEY_SIMPLEX, putText, LINE_AA
from numpy import uint8, zeros
from numpy.typing import NDArray

from pymagewell.events.device_events import (
    TransferCompleteEvent,
    SignalChangeEvent,
    FrameBufferedEvent,
    FrameBufferingEvent,
    TimerEvent,
)
from pymagewell.events.notification import Notification
from pymagewell.pro_capture_device.device_interface import ProCaptureEvents
from pymagewell.pro_capture_device.device_settings import (
    ProCaptureSettings,
    ImageSizeInPixels,
    AspectRatio,
    ImageCoordinateInPixels,
    TransferMode,
)
from pymagewell.pro_capture_device.device_status import (
    TransferStatus,
    SignalStatus,
    FrameInfo,
    OnDeviceBufferStatus,
    FrameState,
    SignalState,
)
from pymagewell.pro_capture_device.pro_capture_device_impl import ProCaptureDeviceImpl

logger = logging.getLogger(__name__)


MOCK_RESOLUTION = ImageSizeInPixels(cols=1920, rows=1080)
MOCK_ASPECT_RATIO = AspectRatio(hor=16, ver=9)
MOCK_FRAME_RATE_HZ = 2

MOCK_FRAME_FONT_SIZE = 1
MOCK_FRAME_LINE_WIDTH = 1
NUM_TEST_FRAMES = 10


class MockProCaptureDevice(ProCaptureDeviceImpl):
    """MockProCaptureDevice is intended to be used during testing, development and CI in the absence of a hardware frame
    grabber or Magwell Windows SDK. Does not require Magewell driver or hardware.

    Only TransferMode.Timer is supported.

    The class generates test frames. The frame rate is limited to 2 frames per second because copying the mock frames
    to a provided PC transfer buffer takes a surprisingly long time (~0.11s)."""

    def __init__(self, settings: ProCaptureSettings):
        """
        Args:
            settings (ProCaptureSettings): The settings to use for the mock device. settings.transfer_mode must be
              set to TransferMode.Timer.
        """
        if settings.transfer_mode != TransferMode.TIMER:
            raise ValueError("MockProCaptureDevice only works in Timer transfer mode.")
        super().__init__(settings)
        self._is_grabbing = False
        self._events = ProCaptureEvents(
            transfer_complete=TransferCompleteEvent(),
            signal_change=SignalChangeEvent(),
            frame_buffered=FrameBufferedEvent(),
            frame_buffering=FrameBufferingEvent(),
            timer_event=TimerEvent(),
        )
        self._events.signal_change.register(Notification(0, 0))
        self._events.timer_event.register(Notification(0, 0))

        self._mock_timer = _MockTimer(self._events.timer_event, MOCK_FRAME_RATE_HZ)

        self._frame_counter: int = 0
        self._mock_frames = [create_mock_frame() for _ in range(NUM_TEST_FRAMES)]
        for i, frame in enumerate(self._mock_frames):
            putText(
                frame,
                str(i),
                (frame.shape[1] // 2, frame.shape[0] // 2),
                FONT_HERSHEY_SIMPLEX,
                1,
                (255, 255, 255),
                1,
                LINE_AA,
            )

    @property
    def events(self) -> ProCaptureEvents:
        """events property
        Returns:
            A `ProCaptureEvents` object containing handles to the events generated by the device during frame grabbing.
        """
        return self._events

    def schedule_timer_event(self) -> None:
        self._mock_timer.schedule_event()

    @property
    def buffer_status(self) -> OnDeviceBufferStatus:
        return OnDeviceBufferStatus(
            buffer_size_in_frames=1,
            num_chunks_in_buffer=1,
            buffering_field_index=1,
            last_buffered_field_index=1,
            last_buffered_frame_index=1,
            num_fully_buffered_frames=1,
            num_chunks_being_buffered=1,
        )

    @property
    def frame_info(self) -> FrameInfo:
        return FrameInfo(
            state=FrameState.BUFFERED,
            interlaced=False,
            segmented=False,
            dimensions=MOCK_RESOLUTION,
            aspect_ratio=MOCK_ASPECT_RATIO,
            buffering_start_time=datetime.now(),
            buffering_complete_time=datetime.now(),
        )

    @property
    def signal_status(self) -> SignalStatus:
        return SignalStatus(
            state=SignalState.LOCKED,
            start_position=ImageCoordinateInPixels(row=0, col=0),
            image_dimensions=MOCK_RESOLUTION,
            total_dimensions=MOCK_RESOLUTION,
            interlaced=False,
            frame_period_s=1 / MOCK_FRAME_RATE_HZ,
            aspect_ratio=MOCK_ASPECT_RATIO,
            segmented=False,
        )

    @property
    def transfer_status(self) -> TransferStatus:
        return TransferStatus(
            whole_frame_transferred=True,
            num_lines_transferred=MOCK_RESOLUTION.rows,
            num_lines_transferred_previously=MOCK_RESOLUTION.rows,
            frame_index=0,
        )

    def start_grabbing(self) -> None:
        self._is_grabbing = True

    def stop_grabbing(self) -> None:
        self._is_grabbing = False

    def start_a_frame_transfer(self, frame_buffer: Array[c_char]) -> datetime:
        """start_a_frame_transfer immediately writes a mock frame to the provided buffer.
        Args:
            frame_buffer (Array[c_char]): The buffer to write the mock frame to.
        Returns:
            The time (datetime.datetime) the frame transfer was completed.
        """
        frame_buffer[: self.frame_properties.size_in_bytes] = self._mock_frames[  # type: ignore
            self._frame_counter % NUM_TEST_FRAMES
        ].tobytes()
        self.events.transfer_complete.set()
        self._frame_counter += 1
        return datetime.now()

    def shutdown(self) -> None:
        self._is_grabbing = False


def create_mock_frame() -> NDArray[uint8]:
    """Creates a mock frame. Used by MockProCaptureDevice."""
    frame = zeros((MOCK_RESOLUTION.rows, MOCK_RESOLUTION.cols, 3), dtype=uint8)

    white_fit_width_radius = MOCK_RESOLUTION.cols // 2
    circle(frame, (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2), white_fit_width_radius, (255, 255, 255))
    putText(
        frame,
        "White",
        (MOCK_RESOLUTION.cols // 2 - white_fit_width_radius, MOCK_RESOLUTION.rows // 2),
        FONT_HERSHEY_SIMPLEX,
        MOCK_FRAME_FONT_SIZE,
        (255, 255, 255),
        MOCK_FRAME_LINE_WIDTH,
        LINE_AA,
    )

    white_fit_height_radius = MOCK_RESOLUTION.rows // 2
    circle(frame, (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2), white_fit_height_radius, (255, 255, 255))
    putText(
        frame,
        "White",
        (MOCK_RESOLUTION.cols // 2 - white_fit_height_radius, MOCK_RESOLUTION.rows // 2),
        FONT_HERSHEY_SIMPLEX,
        MOCK_FRAME_FONT_SIZE,
        (255, 255, 255),
        MOCK_FRAME_LINE_WIDTH,
        LINE_AA,
    )

    ch1_radius = MOCK_RESOLUTION.rows // 4
    circle(frame, (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2), ch1_radius, (255, 0, 0))
    putText(
        frame,
        "ch1",
        (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2 - ch1_radius),
        FONT_HERSHEY_SIMPLEX,
        MOCK_FRAME_FONT_SIZE,
        (255, 0, 0),
        MOCK_FRAME_LINE_WIDTH,
        LINE_AA,
    )

    ch2_radius = MOCK_RESOLUTION.rows // 6
    circle(frame, (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2), ch2_radius, (0, 255, 0))
    putText(
        frame,
        "ch2",
        (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2 - ch2_radius),
        FONT_HERSHEY_SIMPLEX,
        MOCK_FRAME_FONT_SIZE,
        (0, 255, 0),
        MOCK_FRAME_LINE_WIDTH,
        LINE_AA,
    )

    ch3_radius = MOCK_RESOLUTION.rows // 8
    circle(frame, (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2), ch3_radius, (0, 0, 255))
    putText(
        frame,
        "ch3",
        (MOCK_RESOLUTION.cols // 2, MOCK_RESOLUTION.rows // 2 - ch3_radius),
        FONT_HERSHEY_SIMPLEX,
        MOCK_FRAME_FONT_SIZE,
        (0, 0, 255),
        MOCK_FRAME_LINE_WIDTH,
        LINE_AA,
    )

    return frame


class _MockTimer:
    def __init__(self, timer_event: TimerEvent, rate_hz: float):
        self._last_event_time: Optional[float] = None
        self._timer_event_thread = Thread(target=self._wait_and_generate)
        self._rate_hz = rate_hz
        self._timer_event = timer_event
        self._event_counter = 0

    def schedule_event(self) -> None:
        self._timer_event_thread = Thread(target=self._wait_and_generate)
        self._timer_event_thread.start()

    def _wait_and_generate(self) -> None:
        if self._last_event_time is not None:
            time_since_last_event = monotonic() - self._last_event_time
            time_until_next_event = (1.0 / self._rate_hz) - time_since_last_event
            if time_until_next_event > 0:
                sleep(time_until_next_event)
            elif mod(self._event_counter, 10) == 0:
                logger.warning(
                    f"Mock frame rate is {1 / time_since_last_event:.3f} Hz, "
                    f"which is lower than the requested {self._rate_hz} Hz."
                )
        self._last_event_time = monotonic()
        self._timer_event.set()
        self._event_counter += 1



from dataclasses import dataclass
from datetime import datetime

from numpy import uint8, frombuffer
from numpy.typing import NDArray

from pymagewell.pro_capture_device.device_settings import (
    ImageSizeInPixels,
)


@dataclass
class VideoFrameTimestamps:
    buffering_started: datetime
    buffering_complete: datetime
    transfer_started: datetime
    transfer_complete: datetime


@dataclass
class VideoFrame:
    string_buffer: bytes
    dimensions: ImageSizeInPixels
    timestamps: VideoFrameTimestamps

    def as_array(self) -> NDArray[uint8]:
        # frame = cast(NDArray[uint8], imdecode(frombuffer(self.string_buffer, dtype=uint8), IMREAD_COLOR))
        # frame = resize(frame, (self.dimensions.rows, self.dimensions.cols))
        # return frame
        numpy_image = frombuffer(self.string_buffer, dtype=uint8, count=self.dimensions.cols * self.dimensions.rows * 3)
        numpy_image.resize((self.dimensions.rows, self.dimensions.cols, 3))
        return numpy_image



from copy import copy
from ctypes import create_string_buffer
from time import perf_counter
from unittest import TestCase

import pytest

from pymagewell.events.event_base import wait_for_event
from pymagewell.pro_capture_device import ProCaptureDevice
from pymagewell.pro_capture_device.device_interface import ProCaptureDeviceInterface
from pymagewell.pro_capture_device.device_settings import ProCaptureSettings, TransferMode
from pymagewell.pro_capture_device.mock_pro_capture_device import MockProCaptureDevice


@pytest.mark.usefixtures("hardware_mode")
class TestEvents(TestCase):
    def setUp(self) -> None:
        device_settings = ProCaptureSettings()
        device_settings.transfer_mode = TransferMode.TIMER

        assert hasattr(self, "hardware_mode_is_set")
        if self.hardware_mode_is_set:  # type: ignore
            self._device: ProCaptureDeviceInterface = ProCaptureDevice(device_settings)
        else:
            self._device = MockProCaptureDevice(device_settings)

    def tearDown(self) -> None:
        self._device.shutdown()

    def test_transfer_mode(self) -> None:
        self.assertEqual(self._device.transfer_mode, TransferMode.TIMER)

    def test_timer_mode_event_registration(self) -> None:
        events = self._device.events
        self.assertTrue(events.signal_change.is_registered)
        self.assertTrue(events.timer_event.is_registered)

        self.assertFalse(events.frame_buffered.is_registered)
        self.assertFalse(events.frame_buffering.is_registered)

    def test_schedule_timer_event(self) -> None:
        expected_wait_time = self._device.signal_status.frame_period_s

        # First timer event not necessarily occurring after one frame period, so use second
        self._device.schedule_timer_event()
        wait_for_event(self._device.events.timer_event, timeout_ms=1000)

        # Seconds one happens after 1 frame period
        second_frame_start_time = perf_counter()
        self._device.schedule_timer_event()
        wait_for_event(self._device.events.timer_event, timeout_ms=1000)
        time_waited = perf_counter() - second_frame_start_time
        if self.hardware_mode_is_set:  # type: ignore
            self.assertAlmostEqual(time_waited, expected_wait_time, 3)
        else:
            # Mock mode timing is less accurate
            self.assertAlmostEqual(time_waited, expected_wait_time, 1)

    def test_frame_transfer(self) -> None:
        transfer_buffer = create_string_buffer(3840 * 2160 * 4)
        self._device.start_grabbing()
        self._device.schedule_timer_event()
        buffer_before = copy(transfer_buffer)
        wait_for_event(self._device.events.timer_event, timeout_ms=1000)
        self._device.start_a_frame_transfer(transfer_buffer)
        wait_for_event(self._device.events.transfer_complete, timeout_ms=1000)
        buffer_after = copy(transfer_buffer)
        self.assertFalse(buffer_before == buffer_after)



from abc import ABC, abstractmethod
from typing import Optional, Any, List


try:
    import win32api
    import win32event
except ModuleNotFoundError as e:
    try:
        from pymagewell.events import mock_win32api as win32api, mock_win32event as win32event
    except ModuleNotFoundError:
        raise e

from pymagewell.events.notification import Notification
from pymagewell.exceptions import ProCaptureError, WaitForEventTimeout


class Event:
    def __init__(self) -> None:
        self._win32_event = win32event.CreateEvent(None, False, False, None)

    @property
    def win32_event(self) -> Any:
        return self._win32_event

    def set(self) -> None:
        win32event.SetEvent(self.win32_event)

    def destroy(self) -> None:
        win32api.CloseHandle(int(self.win32_event))
        self._win32_event = 0


class RegisterableEvent(Event, ABC):
    def __init__(self) -> None:
        super().__init__()
        self._state: EventState = Unregistered()
        self._notification: Optional[Notification] = None

    @property
    def notification(self) -> Optional[Notification]:
        return self._state.get_notification(self)

    def register(self, notification: Notification) -> None:
        self._state = self._state.register(notification)

    @property
    @abstractmethod
    def registration_token(self) -> Optional[int]:
        raise NotImplementedError()

    @property
    def is_registered(self) -> bool:
        return isinstance(self._state, Registered)


class EventState(ABC):
    @abstractmethod
    def register(self, notification: Notification) -> "EventState":
        raise NotImplementedError()

    @abstractmethod
    def get_notification(self, event: Event) -> Optional[Notification]:
        raise NotImplementedError()


class Registered(EventState):
    def __init__(self, notification: Notification):
        self._notification = notification

    def register(self, notification: Notification) -> "EventState":
        return self

    def get_notification(self, event: Event) -> Optional[Notification]:
        return self._notification


class Unregistered(EventState):
    def register(self, notification: Notification) -> "EventState":
        return Registered(notification)

    def get_notification(self, event: Event) -> Optional[Notification]:
        return None


def wait_for_events(events: List[Event], timeout_ms: int) -> Event:
    result = win32event.WaitForMultipleObjects(tuple([event.win32_event for event in events]), False, timeout_ms)
    if result == 258:
        raise WaitForEventTimeout("Error: wait timed out")
    elif result == win32event.WAIT_OBJECT_0 + 0:
        return events[0]
    elif result == win32event.WAIT_OBJECT_0 + 1:
        return events[1]
    else:
        raise ProCaptureError(f"Wait for event failed: error code {result}")


def wait_for_event(event: Event, timeout_ms: int) -> None:
    result = win32event.WaitForSingleObject(event.win32_event, timeout_ms)
    if result == 258:
        raise WaitForEventTimeout("Error: wait timed out")



import os
import platform
from ctypes import *
from pathlib import Path

MW_SUCCEEDED = 0
MW_FAILED = 1
MW_INVALID_PARAMS = 2


def mw_fourcc(str):
    t_val = 0
    i = 0
    for c in str:
        t_num = ord(c)
        t_val += (t_num<<i)
        i+=8
    return t_val

MWFOURCC_UNK   = mw_fourcc('UNKN')
MWFOURCC_GREY  = mw_fourcc('GREY')
MWFOURCC_Y800  = mw_fourcc('Y800')
MWFOURCC_BGRA  = mw_fourcc('BGRA')
MWFOURCC_Y8    = mw_fourcc('Y8  ')
MWFOURCC_Y16   = mw_fourcc('Y16 ')
MWFOURCC_RGB15 = mw_fourcc('RGB5')
MWFOURCC_RGB16 = mw_fourcc('RGB6')
MWFOURCC_RGB24 = mw_fourcc('RGB ')
MWFOURCC_RGBA  = mw_fourcc('RGBA')
MWFOURCC_ARGB  = mw_fourcc('ARGB')
MWFOURCC_BGR15 = mw_fourcc('BGR5')
MWFOURCC_BGR16 = mw_fourcc('BGR6')
MWFOURCC_BGR24 = mw_fourcc('BGR ')
MWFOURCC_BGRA  = mw_fourcc('BGRA')
MWFOURCC_ABGR  = mw_fourcc('ABGR')
MWFOURCC_NV16  = mw_fourcc('NV16')
MWFOURCC_NV61  = mw_fourcc('NV61')
MWFOURCC_I422  = mw_fourcc('I422')
MWFOURCC_YV16  = mw_fourcc('YV16')
MWFOURCC_YUY2  = mw_fourcc('YUY2')
MWFOURCC_YUYV  = mw_fourcc('YUYV')
MWFOURCC_UYVY  = mw_fourcc('UYVY')
MWFOURCC_YVYU  = mw_fourcc('YVYU')
MWFOURCC_VYUY  = mw_fourcc('VYUY')
MWFOURCC_I420  = mw_fourcc('I420')
MWFOURCC_IYUV  = mw_fourcc('IYUV')
MWFOURCC_NV12  = mw_fourcc('NV12')
MWFOURCC_YV12  = mw_fourcc('YV12')
MWFOURCC_NV21  = mw_fourcc('NV21')
MWFOURCC_P010  = mw_fourcc('P010')
MWFOURCC_P210  = mw_fourcc('P210')
MWFOURCC_IYU2  = mw_fourcc('IYU2')
MWFOURCC_V308  = mw_fourcc('V308')
MWFOURCC_AYUV  = mw_fourcc('AYUV')
MWFOURCC_UYVA  = mw_fourcc('UYVA')
MWFOURCC_V408  = mw_fourcc('v408')
MWFOURCC_VYUA  = mw_fourcc('VYUA')
MWFOURCC_V210  = mw_fourcc('v210')
MWFOURCC_Y410  = mw_fourcc('Y410')
MWFOURCC_V410  = mw_fourcc('v410')
MWFOURCC_RGB10 = mw_fourcc('RG10')
MWFOURCC_BGR10 = mw_fourcc('BG10')

def fourcc_is_packed(fourcc):
    if fourcc == MWFOURCC_NV12:
        return False
    
    if fourcc == MWFOURCC_NV21:
        return False

    if fourcc == MWFOURCC_YV12:
        return False

    if fourcc == MWFOURCC_IYUV:
        return False

    if fourcc == MWFOURCC_I420:
        return False

    if fourcc == MWFOURCC_I422:
        return False

    if fourcc == MWFOURCC_YV16:
        return False

    if fourcc == MWFOURCC_NV16:
        return False

    if fourcc == MWFOURCC_NV61:
        return False

    if fourcc == MWFOURCC_P010:
        return False

    if fourcc == MWFOURCC_P210:
        return False

    return True

def fourcc_get_bpp(fourcc):
    if fourcc == MWFOURCC_GREY:
        return 8
    if fourcc == MWFOURCC_Y800:
        return 8
    if fourcc == MWFOURCC_Y8:
        return 8

    if fourcc == MWFOURCC_I420:
        return 12
    if fourcc == MWFOURCC_IYUV:
        return 12
    if fourcc == MWFOURCC_YV12:
        return 12
    if fourcc == MWFOURCC_NV12:
        return 12
    if fourcc == MWFOURCC_NV21:
        return 12
    
    if fourcc == MWFOURCC_Y16:
        return 16
    if fourcc == MWFOURCC_RGB15:
        return 16
    if fourcc == MWFOURCC_BGR15:
        return 16
    if fourcc == MWFOURCC_RGB16:
        return 16
    if fourcc == MWFOURCC_BGR16:
        return 16
    if fourcc == MWFOURCC_YUY2:
        return 16
    if fourcc == MWFOURCC_YUYV:
        return 16
    if fourcc == MWFOURCC_UYVY:
        return 16
    if fourcc == MWFOURCC_YVYU:
        return 16
    if fourcc == MWFOURCC_VYUY:
        return 16
    if fourcc == MWFOURCC_I422:
        return 16
    if fourcc == MWFOURCC_YV16:
        return 16
    if fourcc == MWFOURCC_NV16:
        return 16
    if fourcc == MWFOURCC_NV61:
        return 16

    if (fourcc == MWFOURCC_IYU2 or 
        fourcc == MWFOURCC_V308 or 
        fourcc == MWFOURCC_RGB24 or 
        fourcc == MWFOURCC_BGR24 or 
        fourcc == MWFOURCC_P010 or 
        fourcc == MWFOURCC_V210):
        return 24
    
    if fourcc == MWFOURCC_AYUV:
        return 32
    if fourcc == MWFOURCC_UYVA:
        return 32
    if fourcc == MWFOURCC_V408:
        return 32
    if fourcc == MWFOURCC_VYUA:
        return 32
    if fourcc == MWFOURCC_RGBA:
        return 32
    if fourcc == MWFOURCC_BGRA:
        return 32
    if fourcc == MWFOURCC_ARGB:
        return 32
    if fourcc == MWFOURCC_ABGR:
        return 32
    if fourcc == MWFOURCC_Y410:
        return 32
    if fourcc == MWFOURCC_V410:
        return 32
    if fourcc == MWFOURCC_P210:
        return 32
    if fourcc == MWFOURCC_RGB10:
        return 32
    if fourcc == MWFOURCC_BGR10:
        return 32
    return 0

def fourcc_calc_min_stride(fourcc,cx,algin):
    t_b_packed = fourcc_is_packed(fourcc)
    if t_b_packed == True:
        if fourcc == MWFOURCC_V210:
            cx = (cx+47)//48*48
            cbline = cx*8//3
        else:
            t_n_bpp = fourcc_get_bpp(fourcc)
            cbline = (cx*t_n_bpp)//8
    else:
        if fourcc == MWFOURCC_P010:
            cbline = cx*2
        elif fourcc == MWFOURCC_P210:
            cbline = cx*2
        else:
            cbline = cx
    return (cbline+algin-1)&~(algin-1)

def fourcc_calc_image_size(fourcc,cx,cy,cbstride):
    t_b_packed = fourcc_is_packed(fourcc)
    if t_b_packed == True:
        t_cb_line = 0
        if fourcc == MWFOURCC_V210:
            cx = (cx+47)//48*48
            t_cb_line = cx*8//3
        else:
            t_n_bpp = fourcc_get_bpp(fourcc)
            t_cb_line = (cx*t_n_bpp)//8
        if cbstride < t_cb_line:
            return 0
        return cbstride*cy
    else:
        if cbstride < cx:
            return 0
        
        if (fourcc == MWFOURCC_NV12 or 
            fourcc == MWFOURCC_NV21 or 
            fourcc == MWFOURCC_YV12 or 
            fourcc == MWFOURCC_IYUV or 
            fourcc == MWFOURCC_I420):
            if cbstride&1!=0 or cy&1 !=0:
                return 0
            return cbstride*cy*3//2
        elif (fourcc == MWFOURCC_I422 or 
              fourcc == MWFOURCC_YV16 or 
              fourcc == MWFOURCC_NV16 or 
              fourcc == MWFOURCC_NV61):
            if cbstride&1 != 0:
                return 0
            return cbstride*cy*2
        elif fourcc == MWFOURCC_P010:
            if cbstride&3!=0 or cy&1 !=0:
                return 0
            return cbstride*cy*3//2
        elif fourcc == MWFOURCC_P210:
            if cbstride & 3 != 0:
                return 0
            return cbstride*cy*2
        else:
            return 0

# sdk
MWCAP_AUDIO_SAMPLES_PER_FRAME = 192
MWCAP_AUDIO_MAX_NUM_CHANNELS = 8

MWCAP_NOTIFY_VIDEO_SIGNAL_CHANGE = 0x0020
MWCAP_NOTIFY_AUDIO_SIGNAL_CHANGE = 0x0040
MWCAP_NOTIFY_AUDIO_FRAME_BUFFERED = 0x1000

MWCAP_VIDEO_FRAME_ID_NEWEST_BUFFERED = -1

MWCAP_AUDIO_CAPTURE_NODE_DEFAULT = 0

'''
typedef void (*VIDEO_CAPTURE_CALLBACK)(BYTE *pbFrame, int cbFrame, UINT64 u64TimeStamp, void* pParam);
'''
mw_video_capture_callback = CFUNCTYPE(None,c_void_p,c_int32,c_uint64,py_object)

'''
typedef void (__stdcall *AUDIO_CAPTURE_STDCALL_CALLBACK)(const BYTE * pbFrame, int cbFrame, UINT64 u64TimeStamp, void* pParam);
'''
mw_audio_capture_callback = CFUNCTYPE(None,c_void_p,c_int32,c_uint64,py_object)

class mw_version(object):
    def __init__(self):
        super().__init__()
        self.m_maj = 0
        self.m_min = 0
        self.m_build = 0

class mw_device_time(object):
    def __init__(self):
        super().__init__()
        self.m_ll_device_time = c_int64(0)

class mw_notify_status(object):
    def __init__(self):
        super().__init__()
        self.m_ll_notify_status = 0

'''
typedef struct _MWCAP_CHANNEL_INFO {		
	WORD											wFamilyID;								///<Product type, refers to #MW_FAMILY_ID
	WORD											wProductID;								///<device ID, refers to  #MWCAP_PRODUCT_ID
	CHAR											chHardwareVersion;						///<Hardware version ID
	BYTE											byFirmwareID;							///<Firmware ID
	DWORD											dwFirmwareVersion;						///<Firmware version
	DWORD											dwDriverVersion;						///<Driver version
	CHAR											szFamilyName[MW_FAMILY_NAME_LEN];		///<Product name
	CHAR											szProductName[MW_PRODUCT_NAME_LEN];		///<Product type
	CHAR											szFirmwareName[MW_FIRMWARE_NAME_LEN];	///<Firmware name
	CHAR											szBoardSerialNo[MW_SERIAL_NO_LEN];		///<Hardware serial number
	BYTE											byBoardIndex;							///<Rotary ID located on the capture card, 0~F.
	BYTE											byChannelIndex;							///<Channel index of the capture card, which starts from 0.
} MWCAP_CHANNEL_INFO;
'''
class mw_cap_channel_info(Structure):
    _pack_ = 1
    _fields_ = [('wFamilyID',c_ushort),
               ('wProductID',c_ushort),
               ('chHardwareVersion',c_char),
               ('byFirmwareID',c_byte),
               ('dwFirmwareVersion',c_ulong),
               ('dwDriverVersion',c_ulong),
               ('szFamilyName',c_char*64),
               ('szProductName',c_char*64),
               ('szFirmwareName',c_char*64),
               ('szBoardSerialNo',c_char*16),
               ('byBoardIndex',c_byte),
               ('byChannelIndex',c_bool)]

'''
typedef enum _MWCAP_VIDEO_SIGNAL_STATE {
	MWCAP_VIDEO_SIGNAL_NONE,																///<No signal
	MWCAP_VIDEO_SIGNAL_UNSUPPORTED,															///<Invalid signal. The capture card detects a signal but can not lock it.
	MWCAP_VIDEO_SIGNAL_LOCKING,																///<Locking signal. The signal is valid, but unlocked.
	MWCAP_VIDEO_SIGNAL_LOCKED																///<Locked signal. The capture card is ready to capture the input signal.
} MWCAP_VIDEO_SIGNAL_STATE;
'''
MWCAP_VIDEO_SIGNAL_NONE = 0
MWCAP_VIDEO_SIGNAL_UNSUPPORTED = 1
MWCAP_VIDEO_SIGNAL_LOCKING = 2
MWCAP_VIDEO_SIGNAL_LOCKED = 3

'''
typedef enum _MWCAP_VIDEO_FRAME_TYPE {
	MWCAP_VIDEO_FRAME_2D							= 0x00,///<2D video frame
	MWCAP_VIDEO_FRAME_3D_TOP_AND_BOTTOM_FULL		= 0x01,///<Top-and-Bottom 3D  video frame at full resolution
	MWCAP_VIDEO_FRAME_3D_TOP_AND_BOTTOM_HALF		= 0x02,///<Top-and-Bottom 3D  video frame at half resolution
	MWCAP_VIDEO_FRAME_3D_SIDE_BY_SIDE_FULL			= 0x03,///<Full side-by-side 3D video frame
	MWCAP_VIDEO_FRAME_3D_SIDE_BY_SIDE_HALF			= 0x04 ///<Half side-by-side 3D video frame
} MWCAP_VIDEO_FRAME_TYPE;
'''
MWCAP_VIDEO_FRAME_2D = 0x00
MWCAP_VIDEO_FRAME_3D_TOP_AND_BOTTOM_FULL = 0x01
MWCAP_VIDEO_FRAME_3D_TOP_AND_BOTTOM_HALF = 0x02
MWCAP_VIDEO_FRAME_3D_SIDE_BY_SIDE_FULL = 0x03
MWCAP_VIDEO_FRAME_3D_SIDE_BY_SIDE_HALF = 0x04

'''
typedef enum _MWCAP_VIDEO_COLOR_FORMAT {         	
	MWCAP_VIDEO_COLOR_FORMAT_UNKNOWN				= 0x00,///<unknown color format
	MWCAP_VIDEO_COLOR_FORMAT_RGB					= 0x01,///<RGB
	MWCAP_VIDEO_COLOR_FORMAT_YUV601					= 0x02,///<YUV601
	MWCAP_VIDEO_COLOR_FORMAT_YUV709					= 0x03,///<YUV709
	MWCAP_VIDEO_COLOR_FORMAT_YUV2020				= 0x04,///<YUV2020
	MWCAP_VIDEO_COLOR_FORMAT_YUV2020C				= 0x05 ///<YUV2020C
} MWCAP_VIDEO_COLOR_FORMAT;	
'''
MWCAP_VIDEO_COLOR_FORMAT_UNKNOWN = 0x00
MWCAP_VIDEO_COLOR_FORMAT_RGB = 0x01
MWCAP_VIDEO_COLOR_FORMAT_YUV601 = 0x02
MWCAP_VIDEO_COLOR_FORMAT_YUV709 = 0x03
MWCAP_VIDEO_COLOR_FORMAT_YUV2020 = 0x04
MWCAP_VIDEO_COLOR_FORMAT_YUV2020C = 0x05

'''
typedef enum _MWCAP_VIDEO_QUANTIZATION_RANGE {			
	MWCAP_VIDEO_QUANTIZATION_UNKNOWN				= 0x00,///<the default quantization range
	MWCAP_VIDEO_QUANTIZATION_FULL					= 0x01,///<Full range, which has 8-bit data. The black-white color range is 0-255/1023/4095/65535.
	MWCAP_VIDEO_QUANTIZATION_LIMITED				= 0x02 ///<Limited range, which has 8-bit data. The black-white color range is 16/64/256/4096-235(240)/940(960)/3760(3840)/60160(61440).
} MWCAP_VIDEO_QUANTIZATION_RANGE;
'''
MWCAP_VIDEO_QUANTIZATION_UNKNOWN = 0x00
MWCAP_VIDEO_QUANTIZATION_FULL = 0x01
MWCAP_VIDEO_QUANTIZATION_LIMITED = 0x02

'''
typedef enum _MWCAP_VIDEO_SATURATION_RANGE {		
	MWCAP_VIDEO_SATURATION_UNKNOWN					= 0x00,///<The default saturation_range range
	MWCAP_VIDEO_SATURATION_FULL						= 0x01,///<Full range, which has 8-bit data. The black-white color range is 0-255/1023/4095/65535
	MWCAP_VIDEO_SATURATION_LIMITED					= 0x02,///<Limited range, which has 8-bit data. The black-white color range is 16/64/256/4096-235(240)/940(960)/3760(3840)/60160(61440)
	MWCAP_VIDEO_SATURATION_EXTENDED_GAMUT			= 0x03 ///<Extended range, which has 8-bit data. The black-white color range is 1/4/16/256-254/1019/4079/65279
} MWCAP_VIDEO_SATURATION_RANGE;	
'''
MWCAP_VIDEO_SATURATION_UNKNOWN = 0x00
MWCAP_VIDEO_SATURATION_FULL = 0x01
MWCAP_VIDEO_SATURATION_LIMITED = 0x02
MWCAP_VIDEO_SATURATION_EXTENDED_GAMUT = 0x03

'''
typedef enum _MWCAP_VIDEO_FRAME_STATE {
	MWCAP_VIDEO_FRAME_STATE_INITIAL,															///<Initial
	MWCAP_VIDEO_FRAME_STATE_F0_BUFFERING,														///<Buffering top subframe
	MWCAP_VIDEO_FRAME_STATE_F1_BUFFERING,														///<Buffering bottom subframe
	MWCAP_VIDEO_FRAME_STATE_BUFFERED															///<Fully bufferred video frame 
} MWCAP_VIDEO_FRAME_STATE;
'''
MWCAP_VIDEO_FRAME_STATE_INITIAL = 0
MWCAP_VIDEO_FRAME_STATE_F0_BUFFERING = 1
MWCAP_VIDEO_FRAME_STATE_F1_BUFFERING = 2
MWCAP_VIDEO_FRAME_STATE_BUFFERED = 3

'''
typedef enum _MWCAP_VIDEO_DEINTERLACE_MODE {
	MWCAP_VIDEO_DEINTERLACE_WEAVE					= 0x00,///<Weave mode
	MWCAP_VIDEO_DEINTERLACE_BLEND					= 0x01,///<Blend mode
	MWCAP_VIDEO_DEINTERLACE_TOP_FIELD				= 0x02,///<Only uses top subframe data
	MWCAP_VIDEO_DEINTERLACE_BOTTOM_FIELD			= 0x03 ///<Only uses bottom subframe data
} MWCAP_VIDEO_DEINTERLACE_MODE;
'''
MWCAP_VIDEO_DEINTERLACE_WEAVE = 0x00
MWCAP_VIDEO_DEINTERLACE_BLEND = 0x01
MWCAP_VIDEO_DEINTERLACE_TOP_FIELD = 0x02
MWCAP_VIDEO_DEINTERLACE_BOTTOM_FIELD = 0x03

'''
typedef enum _MWCAP_VIDEO_ASPECT_RATIO_CONVERT_MODE {
	MWCAP_VIDEO_ASPECT_RATIO_IGNORE					= 0x00,///<Ignore: Ignores the original aspect ratio and stretches to full-screen.
	MWCAP_VIDEO_ASPECT_RATIO_CROPPING				= 0x01,///<Cropping: Expands to full-screen and remove parts of the image when necessary to keep the original aspect ratio.
	MWCAP_VIDEO_ASPECT_RATIO_PADDING				= 0x02 ///<Padding: Fits to screen and add black borders to keep the original aspect ratio.
} MWCAP_VIDEO_ASPECT_RATIO_CONVERT_MODE;
'''
MWCAP_VIDEO_ASPECT_RATIO_IGNORE = 0x00
MWCAP_VIDEO_ASPECT_RATIO_CROPPING = 0x01
MWCAP_VIDEO_ASPECT_RATIO_PADDING = 0x02

'''
typedef enum _MW_FAMILY_ID {
	MW_FAMILY_ID_PRO_CAPTURE					= 0x00,							///<Pro Capture family
	MW_FAMILY_ID_ECO_CAPTURE					= 0x01,							///<Eco Capture family
	MW_FAMILY_ID_USB_CAPTURE					= 0x02							///<USB Capture family
} MW_FAMILY_ID;
'''
MW_FAMILY_ID_PRO_CAPTURE = 0x00
MW_FAMILY_ID_ECO_CAPTURE = 0x01
MW_FAMILY_ID_USB_CAPTURE = 0x02

'''
typedef struct _MWCAP_VIDEO_SIGNAL_STATUS {
	MWCAP_VIDEO_SIGNAL_STATE						state;									///<Defines the accessibility of this video signal
	int												cols;										///<Horizontal start position
	int												rows;										///<Vertical start position
	int												cx;										///<Image width
	int												cy;										///<Image height
	int												cxTotal;								///<Total width
	int												cyTotal;								///<Total height
	BOOLEAN											bInterlaced;							///<Whether the signal is interlaced 
	DWORD											dwFrameDuration;						///<Frame interval of video frame
	int												nAspectX;								///<Width of video ratio
	int												nAspectY;								///<Height of video ratio
	BOOLEAN											bSegmentedFrame;						///<Whether the signal is segmented frame
	MWCAP_VIDEO_FRAME_TYPE							frameType;								///<video frame type
	MWCAP_VIDEO_COLOR_FORMAT						colorFormat;							///<video color format
	MWCAP_VIDEO_QUANTIZATION_RANGE					quantRange;								///<Quantization range
	MWCAP_VIDEO_SATURATION_RANGE					satRange;								///<saturation_range range
} MWCAP_VIDEO_SIGNAL_STATUS;
'''
class mw_video_signal_status(Structure):
    _pack_ = 1
    _fields_ = [('state',c_int32),
                  ('cols',c_int32),
                  ('rows',c_int32),
                  ('cx',c_int32),
                  ('cy',c_int32),
                  ('cxTotal',c_int32),
                  ('cyTotal',c_int32),
                  ('bInterlaced',c_bool),
                  ('dwFrameDuration',c_int32),
                  ('nAspectX',c_int32),
                  ('nAspectY',c_int32),
                  ('bSegmentedFrame',c_bool),
                  ('frameType',c_int32),
                  ('colorFormat',c_int32),
                  ('quantRange',c_int32),
                  ('satRange',c_int32)]

'''
typedef struct _MWCAP_AUDIO_SIGNAL_STATUS {
	WORD											wChannelValid;							///<Valid audio channel mask.The lowest bit indicates whether the 1st and 2nd channels are valid, the second bit indicates whether the 3rd and 4th channels are valid, the third bit indicates whether the 5th and 6th channels are valid, and the fourth bit indicates whether the 7th and 8th channels are valid.
	BOOLEAN											bLPCM;									///<Whether the signal is LPCM
	BYTE											cBitsPerSample;							///<Bit depth of each audio sampling
	DWORD											dwSampleRate;							///<Sample rate
	BOOLEAN											bChannelStatusValid;					///<Whether channel status is valid
	IEC60958_CHANNEL_STATUS							channelStatus;							///<The audio channel status
} MWCAP_AUDIO_SIGNAL_STATUS;
'''
class mw_audio_signal_status(Structure):
    _pack_ = 1
    _fields_ = [('wChannelValid',c_int16),
                ('bLPCM',c_bool),
                ('cBitsPerSample',c_byte),
                ('dwSampleRate',c_int32),
                ('bChannelStatusValid',c_bool),
                ('channelStatus',c_byte*24)]

'''
typedef struct _MWCAP_VIDEO_CAPS {
	DWORD											dwCaps;									///<Capture capability
	WORD											wMaxInputWidth;							///<Max input width
	WORD											wMaxInputHeight;						///<Max input height
	WORD											wMaxOutputWidth;						///<Max output width
	WORD											wMaxOutputHeight;						///<Max output height
} MWCAP_VIDEO_CAPS;
'''
class mw_video_caps(Structure):
    _pack_ = 1
    _fields_ = [('dwCaps',c_int32),
                  ('wMaxInputWidth',c_int16),
                  ('wMaxInputHeight',c_int16),
                  ('wMaxOutputWidth',c_int16),
                  ('wMaxOutputHeight',c_int16)]

'''
typedef struct _MWCAP_VIDEO_CAPTURE_STATUS {
	MWCAP_PTR64										pvContext;																	///<The context of video capture

	BOOLEAN											bPhysicalAddress;															///<Whether to use the physical address to store the capture data 
	union {
		MWCAP_PTR64									pvFrame;																	///<The memory address to store the capture data 
		LARGE_INTEGER								liPhysicalAddress;															///<The physical address to store the capture data
	};

	int												iFrame;																		///<The index of capturing frame
	BOOLEAN											bFrameCompleted;															///<Whether a frame is fully captured
	WORD											cyCompleted;																///<Number of frames captured
	WORD											cyCompletedPrev;															///<Number of frames captured previously
} MWCAP_VIDEO_CAPTURE_STATUS;
'''
class mw_video_capture_status(Structure):
    _pack_ = 1
    _fields_ = [('pvContext',c_int64),           # this need to be experienced
                  ('bPhysicalAddress',c_bool),
                  ('pvFrame',c_uint64),
                  ('iFrame',c_int32),
                  ('bFrameCompleted',c_bool),
                  ('cyCompleted',c_int16),
                  ('cyCompletedPrev',c_int16)]


'''
typedef struct _MWCAP_VIDEO_BUFFER_INFO {
	DWORD											cMaxFrames;									///<Maximum number of frames in on-board cache

	BYTE											iNewestBuffering;							///<The number of the slices being bufferred. A frame of video data may contain multiple slices.
	BYTE											iBufferingFieldIndex;						///<The sequence number of fields being bufferred.

	BYTE											iNewestBuffered;							///<The sequence number of slices the latest bufferred piece.
	BYTE											iBufferedFieldIndex;						///<The sequence number of the latest bufferred field

	BYTE											iNewestBufferedFullFrame;					///<The sequence number of the latest bufferred frame
	DWORD											cBufferedFullFrames;						///<Number of fully bufferred full frames
} MWCAP_VIDEO_BUFFER_INFO;
'''
class mwcap_video_buffer_info(Structure):
    _pack_ = 1
    _fields_ = [('cMaxFrames',c_int32),
                  ('iNewestBuffering',c_byte),
                  ('iBufferingFieldIndex',c_byte),
                  ('iNewestBuffered',c_byte),
                  ('iBufferedFieldIndex',c_byte),
                  ('iNewestBufferedFullFrame',c_byte),
                  ('cBufferedFullFrames',c_int32)]

'''
typedef struct _MWCAP_SMPTE_TIMECODE {
	BYTE 											byFrames;									///<Frames number
	BYTE											bySeconds;									///<Seconds
	BYTE											byMinutes;									///<Minutes
	BYTE											byHours;									///<Hours
} MWCAP_SMPTE_TIMECODE;
'''
class mwcap_smpte_timecode(Structure):
    _pack_ = 1
    _fields_ = [('byFrames',c_byte),
                ('bySeconds',c_byte),
                ('byMinutes',c_byte),
                ('byHours',c_byte)]

'''
typedef struct _MWCAP_VIDEO_FRAME_INFO {
	MWCAP_VIDEO_FRAME_STATE							state;										///<The state of the video framess

	BOOLEAN											bInterlaced;								///<Whether an interlaced signal
	BOOLEAN											bSegmentedFrame;							///<Whether a segmented frame
	BOOLEAN											bTopFieldFirst;								///<Whether the top subframe is in front
	BOOLEAN											bTopFieldInverted;							///<Whether to reverse the top subframe

	int												cx;											///<Width of video frames
	int												cy;											///<Height of video frames
	int												nAspectX;									///<Width of the ratio 
	int												nAspectY;									///<Height of the ratio

	LONGLONG										allFieldStartTimes[2];						///<Start time of capturing top and bottom subframe respectively
	LONGLONG										allFieldBufferedTimes[2];					///<Fully bufferred time of top and bottom frame respectively
	MWCAP_SMPTE_TIMECODE							aSMPTETimeCodes[2];							///<Time code of top and bottom frame respectively
} MWCAP_VIDEO_FRAME_INFO;
'''
class mwcap_video_frame_info(Structure):
    _pack_ = 1
    _fields_ = [('state',c_int32),
                ('bInterlaced',c_bool),
                ('bSegmentedFrame',c_bool),
                ('bTopFieldFirst',c_bool),
                ('bTopFieldInverted',c_bool),
                ('cx',c_int32),
                ('cy',c_int32),
                ('nAspectX',c_int32),
                ('nAspectY',c_int32),
                ('allFieldStartTimes',c_int64*2),
                ('allFieldBufferedTimes',c_int64*2),
                ('aSMPTETimeCodes',c_ubyte*8)] #crnb change from c_ubyte*8

'''
typedef struct _MWCAP_AUDIO_CAPTURE_FRAME {
	DWORD											cFrameCount;																///<Number of bufferred frames
	DWORD											iFrame;																		///<Current frame index
	DWORD											dwSyncCode;																	///<Sync code of audio frame data
	DWORD											dwFlags;																	///<Reserved
	LONGLONG										llTimestamp;																///<The timestamp of audio frame
	DWORD											adwSamples[MWCAP_AUDIO_SAMPLES_PER_FRAME * MWCAP_AUDIO_MAX_NUM_CHANNELS];	///<Audio sample data. Each sample is 32-bit width, and high bit effective. The priority of the path is: Left0, Left1, Left2, Left3, right0, right1, right2, right3
} MWCAP_AUDIO_CAPTURE_FRAME;
'''
class mwcap_audio_capture_frame(Structure):
    _pack_ = 1
    _fields_ = [('cFrameCount',c_int32),
                ('iFrame',c_int32),
                ('dwSyncCode',c_int32),
                ('dwFlags',c_int32),
                ('llTimestamp',c_int64),
                #('adwSamples',c_int32*(192*8))]
                ('adwSamples',c_ubyte*(192*8*4))]

'''
typedef struct _MWCAP_VIDEO_ECO_CAPTURE_OPEN {
	MWCAP_PTR64										hEvent;																///<Handle of capture event

	DWORD											dwFOURCC;															///<Capture format
	WORD											cx;																	///<Width
	WORD											cy;																	///<Height
	LONGLONG										llFrameDuration;													///<Interval, -1 indicates follow format of input source
} MWCAP_VIDEO_ECO_CAPTURE_OPEN;
'''
class mwcap_video_eco_capture_open(Structure):
    _pack_ = 1
    _fields_ = [('hEvent',c_int64),
                ('dwFOURCC',c_int32),
                ('cx',c_int16),
                ('cy',c_int16),
                ('llFrameDuration',c_int64)]

'''
typedef struct _MWCAP_VIDEO_ECO_CAPTURE_FRAME {
	MWCAP_PTR64										pvFrame;															///<The storage address for video capturing
	DWORD											cbFrame;															///<The size of storage for video capturing
	DWORD											cbStride;															///<Width of capture video frame

	BOOLEAN											bBottomUp;															///<Whether to flip
	MWCAP_VIDEO_DEINTERLACE_MODE					deinterlaceMode;													///<DeinterlaceMode

	MWCAP_PTR64										pvContext;															///<Context of ECO 
} MWCAP_VIDEO_ECO_CAPTURE_FRAME;
'''
class mwcap_video_eco_capture_frame(Structure):
    _pack_ = 1
    _fields_ = [('pvFrame',c_int64),
                ('cbFrame',c_int32),
                ('cbStride',c_int32),
                ('bBottomUp',c_bool),
                ('deinterlaceMode',c_int32),
                ('pvContext',c_int64)]

'''
typedef struct _MWCAP_VIDEO_ECO_CAPTURE_STATUS {
	MWCAP_PTR64										pvContext;															///<frame label for DWORD
	MWCAP_PTR64										pvFrame;															///<Frame data address
	LONGLONG										llTimestamp;														///<Timestamp
} MWCAP_VIDEO_ECO_CAPTURE_STATUS;
'''
class mwcap_video_eco_capture_status(Structure):
    _pack_ = 1
    _fields_ = [('pvContext',c_int64),
                ('pvFrame',c_int64),
                ('llTimestamp',c_int64)]


'''
typedef struct tagRECT
{
    LONG    left;
    LONG    top;
    LONG    right;
    LONG    bottom;
} RECT, *PRECT, NEAR *NPRECT, FAR *LPRECT;
'''
class mw_rect(Structure):
    _pack_ = 1
    _fields_ = [('left',c_int32),
                ('top',c_int32),
                ('right',c_int32),
                ('bottom',c_int32)]

class mw_capture(object):
    def __init__(self):
        super().__init__()
        self.m_lib_mw_capture = 0
        t_bits,t_linkage = platform.architecture()
        if platform.system() == "Windows":
            if t_bits == '64bit':
                self.m_lib_path = str(Path(os.path.abspath(__file__)).parent.joinpath("bin/x64/LibMWCapture.dll"))
            elif t_bits == '32bit':
                self.m_lib_path = str(Path(os.path.abspath(__file__)).parent.joinpath("bin/x86/LibMWCapture.dll"))
            else:
                raise AssertionError("ERROR:unsupported architecture - %s\n"%(t_bits))
            try:
                self.m_lib_mw_capture = cdll.LoadLibrary(self.m_lib_path)
            except OSError as identifier:
                raise AssertionError("ERROR:load %s failed - %s\n"%(self.m_lib_path,identifier))
            else:
                self.load_win_funcs()
        elif platform.system() == "Linux":
            raise AssertionError("ERROR:Current Source dosen't support linux")
        else:
            raise AssertionError("ERROR:Current Source doesn't support %s"%(platform.system()))
    
    def load_win_funcs(self):
        self.m_lib_mw_capture.MWOpenChannelByPath.restype= c_void_p
        self.m_lib_mw_capture.MWCloseChannel.argtypes=[c_void_p]
        self.m_lib_mw_capture.MWCreateVideoCapture.restype = c_void_p
        self.m_lib_mw_capture.MWCreateVideoCapture.argtypes=[
            c_void_p,
            c_int32,
            c_int32,
            c_int32,
            c_int32,
            mw_video_capture_callback,
            py_object]
        self.m_lib_mw_capture.MWDestoryVideoCapture.argtypes = [c_void_p]
        self.m_lib_mw_capture.MWCreateAudioCapture.restype = c_void_p
        self.m_lib_mw_capture.MWCreateAudioCapture.argtypes = [
            c_void_p,
            c_int32,
            c_int32,
            c_int16,
            c_int16,
            mw_audio_capture_callback,
            py_object]
        self.m_lib_mw_capture.MWDestoryAudioCapture.argtypes = [c_void_p]
        self.m_lib_mw_capture.MWGetVideoSignalStatus.argtypes = [
            c_void_p,
            POINTER(mw_video_signal_status)]
        self.m_lib_mw_capture.MWGetAudioSignalStatus.argtypes = [
            c_void_p,
            POINTER(mw_audio_signal_status)]
        self.m_lib_mw_capture.MWGetVideoCaps.argtypes = [
            c_void_p,
            POINTER(mw_video_caps)]
        self.m_lib_mw_capture.MWStartVideoCapture.argtypes = [
            c_void_p,
            c_void_p]
        self.m_lib_mw_capture.MWRegisterNotify.restype = c_int64
        self.m_lib_mw_capture.MWRegisterNotify.argtypes = [
            c_void_p,
            c_void_p,
            c_int64]
        self.m_lib_mw_capture.MWRegisterTimer.restype = c_int64
        self.m_lib_mw_capture.MWRegisterTimer.argtypes = [
            c_void_p,
            c_void_p]
        self.m_lib_mw_capture.MWGetDeviceTime.argtypes = [
            c_void_p,
            POINTER(c_int64)]
        self.m_lib_mw_capture.MWPinVideoBuffer.argtypes = [
            c_void_p,
            c_void_p,
            c_int32]
        self.m_lib_mw_capture.MWScheduleTimer.argtypes = [
            c_void_p,
            c_int64,
            c_int64]
        self.m_lib_mw_capture.MWGetVideoBufferInfo.argtypes = [
            c_void_p,
            POINTER(mwcap_video_buffer_info)]
        self.m_lib_mw_capture.MWGetVideoFrameInfo.argtypes = [
            c_void_p,
            c_byte,
            POINTER(mwcap_video_frame_info)]
        self.m_lib_mw_capture.MWCaptureVideoFrameToVirtualAddressEx.argtypes = [
            c_void_p,    # HCHANNEL 						hChannel
            c_int32,    # int							iFrame
            c_void_p,    # LPBYTE							pbFrame
            c_int32,    # DWORD							cbFrame
            c_int32,    # DWORD							cbStride
            c_bool,     # BOOLEAN							bBottomUp
            c_int64,    # MWCAP_PTR64						pvContext
            c_int32,    # DWORD							dwFOURCC
            c_int32,    # int								cx
            c_int32,    # int								cy
            c_int32,    # DWORD							dwProcessSwitchs
            c_int32,    # int								cyParitalNotify
            c_int64,    # HOSD							hOSDImage
            c_void_p,    # const RECT *					pOSDRects
            c_int32,    # int								cOSDRects
            c_int16,    # SHORT							sContrast
            c_int16,    # SHORT							sBrightness
            c_int16,    # SHORT							sSaturation
            c_int16,    # SHORT							sHue
            c_int32,    # MWCAP_VIDEO_DEINTERLACE_MODE			deinterlaceMode
            c_int32,    # MWCAP_VIDEO_ASPECT_RATIO_CONVERT_MODE	aspectRatioConvertMode
            c_void_p,    # const RECT *							pRectSrc
            c_void_p,    # const RECT *							pRectDest
            c_int32,    # int										nAspectX
            c_int32,    # int										nAspectY
            c_int32,    # MWCAP_VIDEO_COLOR_FORMAT				colorFormat
            c_int32,    # MWCAP_VIDEO_QUANTIZATION_RANGE			quantRange
            c_int32     # MWCAP_VIDEO_SATURATION_RANGE			satRange
        ]
        self.m_lib_mw_capture.MWUnpinVideoBuffer.argtypes = [
            c_void_p,
            c_void_p]
        self.m_lib_mw_capture.MWUnregisterNotify.argtypes = [
            c_void_p,
            c_int64]
        self.m_lib_mw_capture.MWUnregisterTimer.argtypes = [
            c_void_p,
            c_int64]
        self.m_lib_mw_capture.MWStopVideoCapture.argtypes = [c_void_p]
        self.m_lib_mw_capture.MWGetChannelInfo.argtypes = [
            c_void_p,
            POINTER(mw_cap_channel_info)]
        self.m_lib_mw_capture.MWGetVideoCaptureStatus.argtypes = [
            c_void_p,
            POINTER(mw_video_capture_status)]
        self.m_lib_mw_capture.MWStartAudioCapture.argtypes = [c_void_p]
        self.m_lib_mw_capture.MWGetNotifyStatus.argtypes = [
            c_void_p,
            c_int64,
            POINTER(c_int64)]
        self.m_lib_mw_capture.MWCaptureAudioFrame.argtypes = [
            c_void_p,
            POINTER(mwcap_audio_capture_frame)]
        self.m_lib_mw_capture.MWStopAudioCapture.argtypes = [c_void_p]
        self.m_lib_mw_capture.MWStartVideoEcoCapture.argtypes = [
            c_void_p,
            POINTER(mwcap_video_eco_capture_open)]
        self.m_lib_mw_capture.MWGetVideoEcoCaptureStatus.argtypes = [
            c_void_p,
            POINTER(mwcap_video_eco_capture_status)]
        self.m_lib_mw_capture.MWCaptureSetVideoEcoFrame.argtypes = [
            c_void_p,
            c_void_p]
        self.m_lib_mw_capture.MWStopVideoEcoCapture.argtypes = [c_void_p]
        

    def mw_capture_init_instance(self):
        if self.m_lib_mw_capture == 0:
            return False
        return (self.m_lib_mw_capture.MWCaptureInitInstance())

    def mw_capture_exit_instance(self):
        if self.m_lib_mw_capture == 0:
            return
        self.m_lib_mw_capture.MWCaptureExitInstance()

    def mw_get_version(self,version):
        if self.m_lib_mw_capture == 0:
            return
        t_maj = c_byte(0)
        t_min = c_byte(0)
        t_build = c_ushort(0)
        t_ret = 0
        t_ret = self.m_lib_mw_capture.MWGetVersion(
            byref(t_maj),
            byref(t_min),
            byref(t_build))
        print("version %d-%d-%d" %(
            t_maj.value,
            t_min.value,
            t_build.value
        ))
        version.m_maj = t_maj.value
        version.m_min = t_maj.value
        version.m_build = t_build.value
        return t_ret

    def mw_refresh_device(self):
        if self.m_lib_mw_capture == 0:
            return
        self.m_lib_mw_capture.MWRefreshDevice()
    
    def mw_get_channel_count(self):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetChannelCount()

    def mw_get_channel_info_by_index(self,index,channelinfo):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetChannelInfoByIndex(index,byref(channelinfo))

    def mw_get_channel_info(self,hchannel,channelinfo):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetChannelInfo(hchannel,byref(channelinfo))

    def mw_get_device_path(self,index,path):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetDevicePath(index,byref(path))

    def mw_open_channel_by_path(self,path):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWOpenChannelByPath(byref(path))

    def mw_close_channel(self,hchannel):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWCloseChannel(hchannel)
    '''
    HANDLE
    LIBMWCAPTURE_API
    MWCreateVideoCapture(
	    HCHANNEL 						hChannel,
	    int								nWidth,
	    int								nHeight,
	    int								nFourcc,
	    int								nFrameDuration,
	    VIDEO_CAPTURE_CALLBACK			callback,
	    void*							pParam
	)
    '''
    def mw_create_video_capture(
        self,
        hchannel,
        t_n_width,
        t_n_height,
        t_n_fourcc,
        t_n_frame_duration,
        video_callback,
        p_param):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWCreateVideoCapture(
            hchannel,
            t_n_width,
            t_n_height,
            t_n_fourcc,
            t_n_frame_duration,
            video_callback,
            p_param)

    def mw_destory_video_capture(self,hvideo):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWDestoryVideoCapture(hvideo)

    '''
    HANDLE
    LIBMWCAPTURE_API
    MWCreateAudioCapture(
	HCHANNEL						hChannel,
	MWCAP_AUDIO_CAPTURE_NODE        captureNode,
	DWORD							dwSamplesPerSec,
	WORD							wBitsPerSample,
	WORD							wChannels,
	AUDIO_CAPTURE_CALLBACK			callback,
	void*							pParam
	);
    '''
    def mw_create_audio_capture(
        self,
        hchannel,
        capture_node,
        samples_per_sec,
        bits_per_sample,
        channels,
        audio_callback,
        p_param):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWCreateAudioCapture(
            hchannel,
            capture_node,
            samples_per_sec,
            bits_per_sample,
            channels,
            audio_callback,
            p_param
        )

    def mw_destory_audio_capture(self,haudio):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWDestoryAudioCapture(haudio)

    def mw_get_video_signal_status(self,hchannel,videosignalstaus):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetVideoSignalStatus(hchannel,byref(videosignalstaus))

    def mw_get_audio_signal_status(self,hchannel,audiosignalstatus):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetAudioSignalStatus(hchannel,byref(audiosignalstatus))

    def mw_get_video_caps(self,hchannel,videocaps):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetVideoCaps(hchannel,byref(videocaps))

    def mw_start_video_capture(self,hchannel,captureevent):
        if self.m_lib_mw_capture == 0:
            return -1
        # return self.m_lib_mw_capture.MWStartVideoCapture(hchannel,captureevent)
        t_ptr = int(captureevent)
        return self.m_lib_mw_capture.MWStartVideoCapture(hchannel,c_void_p(t_ptr))

    def mw_register_notify(self,hchannel,notifyevent,enablebits):
        if self.m_lib_mw_capture == 0:
            return -1
        # return self.m_lib_mw_capture.MWRegisterNotify(hchannel,notifyevent,enablebits)
        t_ptr = int(notifyevent)
        return self.m_lib_mw_capture.MWRegisterNotify(hchannel,c_void_p(t_ptr),enablebits)

    def mw_register_timer(self,hchannel,timerevent):
        if self.m_lib_mw_capture == 0:
            return -1
        # return self.m_lib_mw_capture.MWRegisterTimer(hchannel,timerevent)
        t_ptr = int(timerevent)
        return self.m_lib_mw_capture.MWRegisterTimer(hchannel,c_void_p(t_ptr))

    def mw_get_device_time(self,hchannel,devicetime):
        if self.m_lib_mw_capture == 0:
            return -1
        t_time = c_int64(0)
        t_ret = self.m_lib_mw_capture.MWGetDeviceTime(hchannel,byref(t_time))
        devicetime.m_ll_device_time = t_time
        return t_ret

    def mw_pin_video_buffer(self,hchannel,buffer,framesize):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWPinVideoBuffer(hchannel,buffer,framesize)

    def mw_schedule_timer(self,hchannel,timernotify,expiretime):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWScheduleTimer(hchannel,timernotify,expiretime)

    def mw_get_video_buffer_info(self, hchannel, videobufferinfo):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetVideoBufferInfo(hchannel,byref(videobufferinfo))

    def mw_get_video_frame_info(self, hchannel, t_index, t_videoframeinfo):
        if self.m_lib_mw_capture == 0:
            return -1
        t_ret = self.m_lib_mw_capture.MWGetVideoFrameInfo(hchannel,t_index,t_videoframeinfo)
        return t_ret

    def mw_capture_video_frame_to_virtual_address_ex(
        self,
        hchannel,
        iframe,
        pbframe,
        cbframe,
        cbstride,
        bbottomup,
        pvcontext,
        dwfourcc,
        cx,
        cy,
        dwprocessswitchs,
        cypartialnotify,
        hosdimage,
        posdrects,
        cosdrects,
        scontrast,
        sbrightness,
        ssaturation,
        shue,
        deinterlacemode,
        aspectratioconvertmode,
        prectsrc,
        prectdest,
        naspectx,
        naspecty,
        colorformat,
        quantrange,
        satrange
        ):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWCaptureVideoFrameToVirtualAddressEx(
            hchannel,
            iframe,
            pbframe,
            cbframe,
            cbstride,
            bbottomup,
            pvcontext,
            dwfourcc,
            cx,
            cy,
            dwprocessswitchs,
            cypartialnotify,
            hosdimage,
            posdrects,
            cosdrects,
            scontrast,
            sbrightness,
            ssaturation,
            shue,
            deinterlacemode,
            aspectratioconvertmode,
            prectsrc,
            prectdest,
            naspectx,
            naspecty,
            colorformat,
            quantrange,
            satrange
        )

    def mw_unpin_video_buffer(self,hchannel,buffer):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWUnpinVideoBuffer(hchannel,buffer)

    def mw_unregister_notify(self,hchannel,hnotify):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWUnregisterNotify(hchannel,hnotify)

    def mw_unregister_timer(self,hchannel,htimer):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWUnregisterTimer(hchannel,htimer)

    def mw_stop_video_capture(self,hchannel):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWStopVideoCapture(hchannel)

    def mw_get_video_capture_status(self,hchannel,video_capture_status):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetVideoCaptureStatus(hchannel,byref(video_capture_status))

    def mw_start_audio_capture(self,hchannel):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWStartAudioCapture(hchannel)

    def mw_get_notify_status(self,hchannel,hnotify,status):
        if self.m_lib_mw_capture == 0:
            return -1
        t_ull_status = c_int64(0)
        t_ret = self.m_lib_mw_capture.MWGetNotifyStatus(hchannel,hnotify,byref(t_ull_status))
        status.m_ll_notify_status = t_ull_status.value
        return t_ret

    def mw_capture_audio_frame(self,hchannel,audioframe):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWCaptureAudioFrame(hchannel,byref(audioframe))

    def mw_stop_audio_capture(self,hchannel):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWStopAudioCapture(hchannel)

    def mw_start_video_eco_capture(self,hchannel,ecocaptureopen):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWStartVideoEcoCapture(hchannel,byref(ecocaptureopen))

    def mw_get_video_eco_capture_status(self,hchannel,capturestatus):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWGetVideoEcoCaptureStatus(hchannel,byref(capturestatus))

    def mwcapture_set_video_eco_frame(self,hchannel,frame):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWCaptureSetVideoEcoFrame(hchannel,frame)

    def mw_stop_video_eco_capture(self,hchannel):
        if self.m_lib_mw_capture == 0:
            return -1
        return self.m_lib_mw_capture.MWStopVideoEcoCapture(hchannel)


# Christian Baker, King's College London
# Copyright (c) 2022 School of Biomedical Engineering & Imaging Sciences, King's College London
# Licensed under the MIT. You may obtain a copy at https://opensource.org/licenses/MIT.

from pymagewell.pro_capture_device.pro_capture_device import ProCaptureDevice

__all__ = ["ProCaptureDevice"]



import logging
import time
from ctypes import create_string_buffer, string_at
from datetime import datetime
from functools import singledispatchmethod
from typing import Optional

from pymagewell.events.event_base import wait_for_events, wait_for_event
from pymagewell.events.device_events import (
    SignalChangeEvent,
    Event,
    TimerEvent,
    FrameBufferedEvent,
    FrameBufferingEvent,
)
from pymagewell.pro_capture_device.device_interface import ProCaptureDeviceInterface
from pymagewell.exceptions import ProCaptureError, WaitForEventTimeout
from pymagewell.video_frame import VideoFrame, VideoFrameTimestamps
from pymagewell.pro_capture_device.device_settings import TransferMode

logger = logging.getLogger(__name__)


class ProCaptureController:
    """ProCaptureController controls the transfer of frames from a ProCaptureDevice or MockProCaptureDevice to a PC."""

    def __init__(self, device: ProCaptureDeviceInterface):
        """
        Args:
            device (ProCaptureDeviceInterface): The implementation of ProCaptureDeviceInterface to use for frame
              transfer. ProCaptureDevice or MockProCaptureDevice are both valid implementations.
        """
        self._device = device
        self._transfer_buffer = create_string_buffer(3840 * 2160 * 4)
        self._device.start_grabbing()

    def transfer_when_ready(self, timeout_ms: int = 2000) -> VideoFrame:
        """transfer_when_ready wait for the device to be ready to start transferring, transfers it and returns it.

        This method will block until the frame has been transferred or the timeout has been reached.

        In TransferMode.TIMER and TransferMode.NORMAL, frame transfer will start after a whole frame has been grabbed
          by the device. In TransferMode.LOW_LATENCY, frame transfer will start after a frame has started to be
          buffered onto the device.
        """
        if self._device.transfer_mode == TransferMode.TIMER:
            self._device.schedule_timer_event()
        event = self._wait_for_event(timeout_ms=timeout_ms)
        frame = self._handle_event(event)
        if frame is None:
            return self.transfer_when_ready()
        else:
            return frame

    def _wait_for_event(self, timeout_ms: int) -> Event:
        """Wait for events to be raised by the ProCaptureDevice or Timer, and return the raised event."""
        if self._device.transfer_mode == TransferMode.TIMER:
            grab_event: Event = self._device.events.timer_event
        elif self._device.transfer_mode == TransferMode.NORMAL:
            grab_event = self._device.events.frame_buffered
        elif self._device.transfer_mode == TransferMode.LOW_LATENCY:
            grab_event = self._device.events.frame_buffering
        else:
            raise NotImplementedError("Invalid grab mode.")

        events_to_wait_for = [grab_event, self._device.events.signal_change]
        try:
            event_that_occurred = wait_for_events(events_to_wait_for, timeout_ms=timeout_ms)
        except WaitForEventTimeout as e:
            self.shutdown()
            raise e
        return event_that_occurred

    @singledispatchmethod
    def _handle_event(self, event: Event) -> Optional[VideoFrame]:
        """Handle a raised event, including transferring a frame if the event means one is ready."""
        raise NotImplementedError()

    @_handle_event.register
    def _(self, event: TimerEvent) -> Optional[VideoFrame]:
        """If timer event received, then whole frame is on device. Only subscribed to in TIMER mode. This method
        transfers the frame to a buffer in PC memory, makes a copy, marks the buffer memory as free and then returns
        the copy."""
        transfer_started_timestamp = self._device.start_a_frame_transfer(self._transfer_buffer)
        buffering_started_timestamp = self._device.frame_info.buffering_start_time
        transfer_complete_timestamp = self._wait_for_frame_or_chunk_transfer_to_complete(timeout_ms=2000)
        buffering_complete_timestamp = self._device.frame_info.buffering_complete_time
        if not self._device.transfer_status.whole_frame_transferred:  # this marks the buffer memory as free
            raise ProCaptureError("Only part of frame has been acquired")
        return self._format_frame(
            timestamps=VideoFrameTimestamps(
                transfer_started=transfer_started_timestamp,
                transfer_complete=transfer_complete_timestamp,
                buffering_started=buffering_started_timestamp,
                buffering_complete=buffering_complete_timestamp,
            )
        )

    @_handle_event.register
    def _(self, event: FrameBufferedEvent) -> Optional[VideoFrame]:
        """If FrameBufferedEvent event received, then whole frame is on device. This event is only subscribed to in
        NORMAL mode. This method transfers it to a buffer in PC memory, makes a copy, marks the buffer memory as free
        and then returns the copy."""
        transfer_started_timestamp = self._device.start_a_frame_transfer(self._transfer_buffer)
        transfer_complete_timestamp = self._wait_for_frame_or_chunk_transfer_to_complete(timeout_ms=2000)
        buffering_started_timestamp = self._device.frame_info.buffering_start_time
        buffering_complete_timestamp = self._device.frame_info.buffering_complete_time
        if not self._device.transfer_status.whole_frame_transferred:  # this marks the buffer memory as free
            raise ProCaptureError("Only part of frame has been acquired")
        return self._format_frame(
            timestamps=VideoFrameTimestamps(
                transfer_started=transfer_started_timestamp,
                transfer_complete=transfer_complete_timestamp,
                buffering_started=buffering_started_timestamp,
                buffering_complete=buffering_complete_timestamp,
            )
        )

    @_handle_event.register
    def _(self, event: FrameBufferingEvent) -> Optional[VideoFrame]:
        """If FrameBufferingEvent event received, then a frame has started to be acquired by the card. This event is
        only subscribed to in LOW_LATENCY mode. This method starts the transfer of the available lines to a buffer in
        PC memory while the acquisition is still happening. It then waits until all lines have been received (this query
        also frees the memory), copies the buffer contents and returns the copy."""
        transfer_started_timestamp = self._device.start_a_frame_transfer(self._transfer_buffer)
        buffering_started_timestamp = self._device.frame_info.buffering_start_time
        self._wait_for_frame_or_chunk_transfer_to_complete(timeout_ms=2000)
        wait_start_t = time.perf_counter()
        while (
            self._device.transfer_status.num_lines_transferred < self._device.frame_properties.dimensions.rows
            and (time.perf_counter() - wait_start_t) < 1
        ):
            pass
        transfer_complete_timestamp = datetime.now()
        buffering_complete_timestamp = self._device.frame_info.buffering_complete_time

        return self._format_frame(
            timestamps=VideoFrameTimestamps(
                transfer_started=transfer_started_timestamp,
                transfer_complete=transfer_complete_timestamp,
                buffering_started=buffering_started_timestamp,
                buffering_complete=buffering_complete_timestamp,
            )
        )

    @_handle_event.register
    def _(self, event: SignalChangeEvent) -> None:
        """If a SignalChangeEvent is received, then the source signal has changed and no frame is available."""
        logger.info("Frame grabber signal change detected")

    def _format_frame(self, timestamps: VideoFrameTimestamps) -> VideoFrame:
        """Copy the contents of the transfer buffer, and return it as a VideoFrame."""
        # Copy the acquired frame
        string_buffer = string_at(self._transfer_buffer, self._device.frame_properties.size_in_bytes)
        frame = VideoFrame(
            string_buffer,
            dimensions=self._device.frame_properties.dimensions,
            timestamps=timestamps,
        )
        return frame

    def _wait_for_frame_or_chunk_transfer_to_complete(self, timeout_ms: int) -> datetime:
        """Waits until a whole frame (or chunk of a frame in low latency mode) has been transferred to the buffer in
        PC memory."""
        try:
            wait_for_event(self._device.events.transfer_complete, timeout_ms=timeout_ms)
            return datetime.now()
        except WaitForEventTimeout as e:
            self.shutdown()
            raise e

    def shutdown(self) -> None:
        """Shuts down the frame grabber device."""
        self._device.stop_grabbing()
        self._device.shutdown()



# content of conftest.py
import logging
from typing import Any

import pytest

logger = logging.getLogger(__name__)


def pytest_addoption(parser: Any) -> None:
    parser.addoption(
        "--hardware_mode",
        action="store",
        default="True",
        help="Set to true to run tests on hardware (requires video signal)",
    )


@pytest.fixture
def hardware_mode(request: Any) -> None:
    try:
        if request.config.getoption("--hardware") == "True":
            logger.info("Hardware test mode enabled")
            request.cls.hardware_mode_is_set = True
        elif request.config.getoption("--hardware") == "False":
            logger.info("Hardware test mode disabled")
            request.cls.hardware_mode_is_set = False
        else:
            raise ValueError("hardware_mode option must be set to either 'True' or 'False'")
    except ValueError:
        logger.info("Hardware test mode disabled")
        request.cls.hardware_mode_is_set = False
